"""
extended types

>>> a = 5
>>> isinstance(a, integer_types)
True
>>> import numpy as np
>>> a = np.int32(7)
>>> isinstance(a, int)
False
>>> isinstance(a, integer_types)
True
"""

import numpy as np

integer_types = (int, np.int32, np.int64)
float_types = (float, np.float32, np.float64)

if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



import logging
from collections import defaultdict, namedtuple
from functools import cached_property
from pathlib import Path

import numpy as np

from nastranio.readers.gmsh.mixins import Modificator

Entity = namedtuple("Entity", ["entityDim", "entityTag", "bbox", "physical_tags"])
PhysicalGroup = namedtuple("PhysicalGroup", ["dim", "physical_tag", "name"])
Node = namedtuple("Node", ["gid", "entity", "x", "y", "z"])
Element = namedtuple(
    "Element",
    [
        "eid",
        "gids",
        "entity_dim",
        "entity_tag",
        "element_type",
        "physical_tags",
        "physical_groups",
    ],
)


def revert_dict(d):
    newd = defaultdict(set)
    # if `d` is kind of {k: (iterable)}
    if isinstance(next(iter(d.values())), (tuple, list, set)):
        for k, v in d.items():
            for subv in v:
                newd[subv].add(k)
    else:
        for k, v in d.items():
            newd[v] = k
    return dict(newd)


class GmsHParser(Modificator):
    def __init__(self, filepath):
        self.read(filepath)

    def info(self):
        msg = ""
        msg += f"Mesh Format {self._mesh_format}\n"
        msg += f"nb. nodes: {len(self.nodes)}\n"
        msg += f"nb. elements: {len(self.elements)}\n"
        cnt_types = {k: len(v) for k, v in self.type2eid.items()}
        for typ, nb in cnt_types.items():
            msg += f"   Type {typ}: {nb} elements\n"
        coincident_gids = self.coincident_nodes()
        if coincident_gids:
            msg += f"Coincident gids: {coincident_gids}\n"
        return msg

    def coincident_nodes(self, digits=0):
        xyzs = np.array([[n.gid, n.x, n.y, n.z] for gid, n in self.nodes.items()])
        gids = xyzs[:, 0].astype(int)
        xyzs = xyzs[:, 1:]
        if digits:
            xyzs = np.round(xyzs, digits)
        dedup_values, count = np.unique(xyzs, axis=0, return_counts=True)
        ix2gid = dict(zip(range(len(gids)), gids))
        repeated_groups = dedup_values[count > 1]
        coincident_gids = []
        for repeated_group in repeated_groups:
            repeated_idx = np.argwhere(np.all(xyzs == repeated_group, axis=1)).ravel()
            repeated_idx = (ix2gid[ix] for ix in repeated_idx)
            coincident_gids.append(tuple(repeated_idx))
        return coincident_gids

    def read(self, filepath):
        with open(filepath) as fh:
            self._toc = {}
            self._build_toc(fh.readlines())
        for section in self._toc.keys():
            process_func = f"_process_{section}"
            if hasattr(self, process_func):
                getattr(self, process_func)()
            else:
                logging.warning(f"no processing function for {section}")
        self._make_gid_eid_type()
        self._make_eid_gid_physicalGroups()

    # =========================================================================
    # a few helpers
    # =========================================================================
    # @cached_property
    # def type2eid(self):
    #     eid2type = defaultdict(set)
    #     for eid, elt in self.elements.items():
    #         eid2type[elt.element_type].add(eid)
    #     return dict(eid2type)

    @cached_property
    def dim2eid(self):
        eid2type = defaultdict(set)
        for eid, elt in self.elements.items():
            eid2type[elt.entity_dim].add(eid)
        return dict(eid2type)

    @cached_property
    def dim2physical_group(self):
        dim2pg = defaultdict(set)
        for (dim, _id), pg in self.physical_groups.items():
            dim2pg[pg.dim].add(pg)
        return dict(dim2pg)

    def _make_gid_eid_type(self):
        """build:

        * self.eid2gids
        * self.gid2eids
        * self.type2eid
        * self.eid2type

        """
        eid2gids = {}
        gid2eids = defaultdict(set)
        type2eid = defaultdict(set)
        eid2type = {}
        for eid, elt in self.elements.items():
            eid2gids[eid] = set(elt.gids)
            type2eid[elt.element_type].add(eid)
            eid2type[eid] = elt.element_type
            for gid in elt.gids:
                gid2eids[gid].add(eid)
        self.eid2gids = eid2gids
        self.type2eid = dict(type2eid)
        self.gid2eids = dict(gid2eids)
        self.eid2type = eid2type

    def _make_eid_gid_physicalGroups(self):
        """build:

        * self.eid2physical_groups
        * self.physical_group2eids
        * self.physical_group2gids
        * self.gid2physical_groups

        """
        self.eid2physical_groups = {
            eid: set(elt.physical_groups) for eid, elt in self.elements.items()
        }
        physical_group2eids = defaultdict(set)
        gid2physical_groups = defaultdict(set)
        physical_group2gids = defaultdict(set)
        for eid, grps in self.eid2physical_groups.items():
            gids = self.eid2gids[eid]
            for grp in grps:
                physical_group2eids[grp].add(eid)
                physical_group2gids[grp] |= gids
            for gid in gids:
                gid2physical_groups[gid] |= grps
        self.physical_group2eids = dict(physical_group2eids)
        self.physical_group2gids = dict(physical_group2gids)
        self.gid2physical_groups = dict(gid2physical_groups)

    # =========================================================================
    # internal stuff
    # =========================================================================

    def _build_toc(self, content):
        buffer = []
        IN_SECTION = None
        for line in content:
            line = line.split("//")[0].strip()
            if line.startswith("//"):
                continue
            if line.startswith("$End"):
                assert line[4:] == IN_SECTION
                self._toc[IN_SECTION] = buffer[:]
                buffer = []
                continue
            if line.startswith("$"):
                IN_SECTION = line[1:]
                continue
            buffer.append(line)

    def _process_MeshFormat(self):
        """creates self.mesh_format"""
        self._mesh_format = self._toc["MeshFormat"][0].split()[0]

    def _process_PhysicalNames(self):
        """
        creates `self.physical_groups` mapping (dim, physical tag) to physicalGroup named tuple
        """
        content = self._toc["PhysicalNames"]
        nb_physical_names = int(content.pop(0))
        data = {}
        for line in content:
            dim, physical_tag, *name = line.split()
            dim = int(dim)
            name = " ".join(name)
            _ = PhysicalGroup(dim, int(physical_tag), name.strip('"'))
            data[(dim, _.physical_tag)] = _
        self.physical_groups = data

    def _process_Entities_1d2d3d(self, nb, content, dim):
        for _ in range(nb):
            itemdata = content.pop(0)
            if dim == 0:
                entityTag, minX, minY, minZ, *groups = itemdata.split()
                bbox = (minX, minY, minZ)
            else:
                (
                    entityTag,
                    minX,
                    minY,
                    minZ,
                    maxX,
                    maxY,
                    maxZ,
                    *groups,
                ) = itemdata.split()
                bbox = (minX, minY, minZ, maxX, maxY, maxZ)
            entityTag = int(entityTag)
            # processing numPhysicalTags
            numPhysicalTags = int(groups.pop(0))
            physicalTags = groups[:numPhysicalTags]
            assert numPhysicalTags == len(physicalTags)
            groups = groups[numPhysicalTags:]
            if dim == 0:
                assert len(groups) == 0
            else:
                # should be finished here with 0d
                numBounding = int(groups[0])
                pointTags = list(map(int, groups[1:]))
                assert len(pointTags) == numBounding
            entity = Entity(
                entityDim=dim,
                entityTag=entityTag,
                bbox=tuple(map(float, bbox)),
                physical_tags=tuple(map(int, physicalTags)),
            )
            self.entities[dim][entityTag] = entity
        return content

    def _process_Entities(self):
        """
        creates a `self.entities` dictionnary, mapping
        {0: {entityTag: {"bbox": (0, 0, 0), "physicalTags": []}}}
        """
        content = self._toc["Entities"]
        self.entities = {k: defaultdict(dict) for k in range(4)}
        nb = content.pop(0).split()
        # ----------------------------------------------------------------
        # processing points
        numPoints, numCurves, numSurfaces, numVolumes = map(int, nb)
        content = self._process_Entities_1d2d3d(numPoints, content, 0)
        content = self._process_Entities_1d2d3d(numCurves, content, 1)
        content = self._process_Entities_1d2d3d(numSurfaces, content, 2)
        content = self._process_Entities_1d2d3d(numVolumes, content, 3)
        self.entities = {dim: dict(data) for dim, data in self.entities.items()}

    def _process_Nodes(self):
        content = self._toc["Nodes"]
        numEntityBlocks, numNodes, minNodeTag, maxNodeTag = map(
            int, content.pop(0).split()
        )
        self.nodes = {}

        for blockno in range(numEntityBlocks):
            entityDim, entityTag, parametric, numNodesInBlock = map(
                int, content.pop(0).split()
            )
            if numNodesInBlock == 0:
                continue
            node_ids = tuple(map(int, ",".join(content[:numNodesInBlock]).split(",")))
            xyzs = "\n".join(content[numNodesInBlock : numNodesInBlock * 2])
            content = content[2 * numNodesInBlock :]
            xyzs = np.fromstring(xyzs, dtype=float, sep=" ")
            xyzs = xyzs.reshape(int(len(xyzs) / 3), 3)
            for i, gid in enumerate(node_ids):
                xyz = xyzs[i]
                node = Node(
                    gid=gid,
                    entity=self.entities[entityDim][entityTag],
                    x=xyz[0],
                    y=xyz[1],
                    z=xyz[2],
                )
                self.nodes[gid] = node

    def _process_Elements(self):
        self.elements = {}
        # ---------------------------------------------------------------------
        # parse content
        content = self._toc["Elements"]
        numEntityBlocks, numElements, minElementTag, maxElementTag = map(
            int, content.pop(0).split()
        )

        for blockno in range(numEntityBlocks):
            entityDim, entityTag, elementType, numElementsInBlock = map(
                int, content.pop(0).split()
            )
            physicalTags = tuple(self.entities[entityDim][entityTag].physical_tags)
            try:
                physicalNames = tuple(
                    (
                        self.physical_groups[(entityDim, id_)].name
                        for id_ in physicalTags
                    )
                )
            except KeyError:
                # GMSH V4.12 API changes???
                physicalNames = tuple(
                    (
                        self.physical_groups[(entityDim, -id_)].name
                        for id_ in physicalTags
                    )
                )
            buffer = content[:numElementsInBlock]
            content = content[numElementsInBlock:]
            for row in buffer:
                eid_and_gids = map(int, row.split())
                eid, *gids = eid_and_gids
                element = Element(
                    eid=eid,
                    gids=tuple(gids),
                    entity_dim=entityDim,
                    entity_tag=entityTag,
                    element_type=elementType,
                    physical_tags=physicalTags,
                    physical_groups=physicalNames,
                )
                self.elements[eid] = element

    if __name__ == "__main__":
        import doctest

        doctest.testmod(optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE)



"""
NASTRAN material Cards Collection
"""
import logging
import re
from collections import defaultdict

from nastranio.cardslib import SimpleCard
from nastranio.decorators import material


@material
class MAT1(SimpleCard):
    """
    Isotropic Material Property Definition
    Defines the material properties for linear isotropic materials.

    ref: NX Nastran 12 Quick Reference Guide 15-2 (p.1968)
    """

    COMMENTS_KEY = "Material"

    TABLE = """
    | 1    | 2   | 3  | 4  | 5     | 6   | 7 | 8    | 9  | 10 |
    |------+-----+----+----+-------+-----+---+------+----+----|
    | MAT1 | MID | E  | G  | NU    | RHO | A | TREF | GE |    |
    |      | ST  | SC | SS | MCSID |     |   |      |    |    |
    """

    DEFAULTS = {
        "G": None,
        "NU": None,
        "RHO": 0.0,
        "A": 0.0,
        "TREF": 0.0,
        "GE": None,
        "ST": None,
        "SC": None,
        "SS": None,
        "MCSID": None,
    }


@material
class MAT8(SimpleCard):
    """
    Shell Element Orthotropic Material Property Definition
    Defines the material property for an orthotropic material for isoparametric shell
    elements.

    ref: NX Nastran 12 Quick Reference Guide 15-19 (p.1985)
    """

    COMMENTS_KEY = "Material"

    TABLE = """
    | 1    | 2   | 3   | 4    | 5    | 6   | 7   | 8   | 9   | 10 |
    |------+-----+-----+------+------+-----+-----+-----+-----+----|
    | MAT8 | MID | E1  | E2   | NU12 | G12 | G1Z | G2Z | RHO |    |
    |      | A1  | A2  | TREF | Xt   | Xc  | Yt  | Yc  | S   |    |
    |      | GE  | F12 | STRN |      |     |     |     |     |    |

    """



"""
NastraIO decorators
"""
import inspect
import logging
import re
import time
import warnings
from collections import defaultdict
from functools import lru_cache, partial, wraps
from itertools import chain
from threading import RLock

import numpy as np
import pandas as pd

try:
    from pyinstrument import Profiler

    IS_PYINSTRUMENT = True
except ImportError:
    IS_PYINSTRUMENT = False

from numtools.vgextended import loc_array

from nastranio.constants import (
    AXIS,
    BOUNDARY,
    CARDS_REGISTER,
    ELEMENT,
    LOADING,
    MATERIAL,
    PROPERTY,
    UNKNOWN,
    GMSHElementTypes,
    VTKShapes,
)

_NOT_FOUND = object()


CACHES = defaultdict(list)


# from nastranio.constants import ELEMENT, MATERIAL, AXIS, UNKNOWN, LOADING, PROPERTY, BOUNDARIES
class cached_property:
    """backported from python3.8; modify cached name attribute from "<name>" to
    "{CACHED_PREFIX}{name}" to easily clear cache from the class itself
    """

    CACHED_PREFIX = "_cached_"

    def __init__(self, func):
        self.func = func
        self.attrname = None
        self.__doc__ = func.__doc__
        self.lock = RLock()

    def __set_name__(self, owner, name):
        if self.attrname is None:
            self.attrname = f"{self.CACHED_PREFIX}{name}"
        elif name != self.attrname:
            raise TypeError(
                "Cannot assign the same cached_property to two different names "
                f"({self.attrname!r} and {name!r})."
            )

    def __get__(self, instance, owner=None):
        if instance is None:
            return self

        if self.attrname is None:
            raise TypeError(
                "Cannot use cached_property instance without calling __set_name__ on it."
            )
        try:
            cache = instance.__dict__
        except (
            AttributeError
        ):  # not all objects have __dict__ (e.g. class defines slots)
            msg = (
                f"No '__dict__' attribute on {type(instance).__name__!r} "
                f"instance to cache {self.attrname!r} property."
            )
            raise TypeError(msg) from None
        val = cache.get(self.attrname, _NOT_FOUND)
        if val is _NOT_FOUND:
            with self.lock:
                # check if another thread filled cache while we awaited lock
                val = cache.get(self.attrname, _NOT_FOUND)
                if val is _NOT_FOUND:
                    val = self.func(instance)
                    try:
                        cache[self.attrname] = val
                    except TypeError:
                        msg = (
                            f"The '__dict__' attribute on {type(instance).__name__!r} instance "
                            f"does not support item assignment for caching {self.attrname!r} property."
                        )
                        raise TypeError(msg) from None
        return val


def timeit(method=None, loglevel="info"):
    if not method:
        return partial(timeit, loglevel=loglevel)

    @wraps(method)
    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        delta = te - ts
        name = method.__name__
        msg = f'function or method "{name}" took: {delta:.2f} sec.'
        if not loglevel:
            print(msg)
        else:
            getattr(logging, loglevel)(msg)
        return result

    return timed


def profile(f):
    @wraps(f)
    def wrap(*args, **kw):
        if not IS_PYINSTRUMENT:
            return f(*args, **kw)
        profiler = Profiler()
        profiler.start()
        result = f(*args, **kw)
        profiler.stop()
        name = f.__name__
        print(80 * "=")
        print('Profiling "%s"' % name)
        print(profiler.output_text(unicode=True, color=True))
        print(80 * "=")
        return result

    return wrap


def dump_args(decorated_function):
    """
    Function decorator logging entry + exit and parameters of functions.

    Entry and exit as logging.info, parameters as logging.DEBUG.
    """

    @wraps(decorated_function)
    def wrapper(*dec_fn_args, **dec_fn_kwargs):
        # Log function entry
        func_name = decorated_function.__name__
        log = logging.getLogger(func_name)
        log.debug("Entering {}()...".format(func_name))

        # get function params (args and kwargs)
        argnames = decorated_function.__code__.co_varnames
        if "self" in argnames:
            ix = argnames.index("self")
            clsname = dec_fn_args[ix].__class__.__name__
            log.info("method of %s" % clsname)

        args = {
            k: v for k, v in dict(zip(argnames, dec_fn_args)).items() if k != "self"
        }
        params = dict(args=args, kwargs=dec_fn_kwargs)

        log.info(
            "\t"
            + ", ".join(["{}={}".format(str(k), repr(v)) for k, v in params.items()])
        )
        # Execute wrapped (decorated) function:
        out = decorated_function(*dec_fn_args, **dec_fn_kwargs)
        log.debug("Done running {}()!".format(func_name))

        return out

    return wrapper


# Cards class decorators
# It difines at least (e.g.) unknown:
#    * cls.type
# ========================================================================
def unknown(cls):
    setattr(cls, "type", UNKNOWN)
    return cls


# ------------------------------------------------------------------------
# elements


def nb_nodes(self):
    return len(self.gids_header)


def cells(self, nasgids=None):
    """
    return element's cell definition.
    """
    gidnames = self.gids_header[:]  # ['G1', 'G2', 'G3']
    cells = []
    eids = self.carddata["main"][self.EID_FIELDNAME]
    # ========================================================================
    # default nasgids and naseids
    # ========================================================================
    if nasgids is None:
        nasgids = list(self._eid2gids.values())
        nasgids = np.array(sorted(list(chain(*nasgids))))
        logging.warning("no gids nor eids passed as reference")

    # append number of nodes
    cells.append(np.array(len(self) * [self.nb_nodes()]))
    gidnames.insert(0, "nbpoints")
    # ------------------------------------------------------------------------
    # map VTK nodes to NASTRAN nodes
    for gname in gidnames:
        if gname == "nbpoints":
            continue
        gids = self.carddata["main"][gname]
        gix = loc_array(nasgids, gids)
        cells.append(gix)

    payload = {"data": np.array(cells).T, "index": np.array(eids), "columns": gidnames}
    return payload


def to_vtk(self, nasgids, debug=True):
    """return data to build VTU unstructured file"""
    payload = self.cells(nasgids=nasgids)
    # len(payload['index'])== len(self) for most of the cards
    # except RBEs
    cell_types = len(payload["index"]) * [VTKShapes[self.shape]]
    cells = payload["data"].reshape(-1)
    eids = payload["index"].tolist()
    # offset = np.array([0, 5, ...])
    offset = np.arange(0, len(cells), self.nb_nodes() + 1)
    return {
        "cell_types": cell_types,
        "cells": cells,
        "eids": eids,
        "card": len(cell_types) * [self.card],
        "offset": offset,
    }


def eid2gids(self, keep_order=False, asdf=False):
    """return a dictionnary {'eid' <int>: 'gids': <set>} based on MAIN CARD data

    Also eventually call a `eid2gids_complement()` hook for cards having additional data
    to provide (eg. RBE2/RBE3).
    """
    if keep_order or asdf is True:
        eid2gids = self._eid2gids_ordered
    else:
        eid2gids = self._eid2gids
    if asdf:
        df = pd.DataFrame(eid2gids).T
        df.index.names = ["EID"]
        df.columns = self.gids_header
        return df

    return dict(eid2gids)


def update_gid(self, eid, gidno, new_gid):
    data = self.carddata["main"]
    eid_ix = data["EID"].index(eid)
    header = self.gids_header[gidno]  # -> eg. "GB"
    old_gid = data[header][eid_ix]
    data[header][eid_ix] = new_gid
    return eid_ix, old_gid


def element(dim, shape):
    def decorator(cls):
        CARDS_REGISTER.append(cls.__name__)
        setattr(cls, "type", ELEMENT)
        setattr(cls, "dim", dim)
        setattr(cls, "shape", shape)
        setattr(cls, "gmsh_eltype", GMSHElementTypes.get(shape))
        if not hasattr(cls, "eid2gids"):
            setattr(cls, "eid2gids", eid2gids)
        if not hasattr(cls, "cells"):
            setattr(cls, "cells", cells)
        if not hasattr(cls, "nb_nodes"):
            setattr(cls, "nb_nodes", nb_nodes)
        if not hasattr(cls, "update_gid"):
            setattr(cls, "update_gid", update_gid)
        if not hasattr(cls, "to_vtk"):
            setattr(cls, "to_vtk", to_vtk)
        if not hasattr(cls, "GIDS_PATTERN"):
            setattr(cls, "GIDS_PATTERN", re.compile(r"^G\d+$"))
        if not hasattr(cls, "THK_PATTERN"):
            setattr(cls, "THK_PATTERN", re.compile(r"^T\d+$"))
        if not hasattr(cls, "EID_FIELDNAME"):
            setattr(cls, "EID_FIELDNAME", "EID")
        if not hasattr(cls, "PID_FIELDNAME") and "PID" in cls.TABLE:
            setattr(cls, "PID_FIELDNAME", "PID")
        if not hasattr(cls, "XID_FIELDNAME"):
            setattr(cls, "XID_FIELDNAME", getattr(cls, "EID_FIELDNAME"))
        return cls

    return decorator


def fem_property(cls):
    CARDS_REGISTER.append(cls.__name__)
    setattr(cls, "type", PROPERTY)
    if not hasattr(cls, "PID_FIELDNAME") and "PID" in cls.TABLE:
        setattr(cls, "PID_FIELDNAME", "PID")
    if not hasattr(cls, "MATS_PATTERN"):
        setattr(cls, "MATS_PATTERN", re.compile(r"^MID\d*$"))
    if not hasattr(cls, "XID_FIELDNAME"):
        setattr(cls, "XID_FIELDNAME", getattr(cls, "PID_FIELDNAME"))
    return cls


def axis(cls):
    CARDS_REGISTER.append(cls.__name__)
    setattr(cls, "type", AXIS)
    return cls


def loading_type(typ=None):
    def decorator(cls):
        CARDS_REGISTER.append(cls.__name__)
        setattr(cls, "type", LOADING)
        if not hasattr(cls, "XID_FIELDNAME"):
            setattr(cls, "XID_FIELDNAME", "SID")
        if typ:
            if typ == "nodal":
                setattr(cls, "LOADING_TYPE", "nodes")
            elif typ == "elemental":
                setattr(cls, "LOADING_TYPE", "elements")
            else:
                raise ValueError(f"decorator 'loading_type' does not handle {typ=}")
        else:
            setattr(cls, "LOADING_TYPE", None)
        return cls

    return decorator


def boundary(cls):
    CARDS_REGISTER.append(cls.__name__)
    setattr(cls, "type", BOUNDARY)
    if not hasattr(cls, "XID_FIELDNAME"):
        setattr(cls, "XID_FIELDNAME", "SID")
    return cls


def material(cls):
    CARDS_REGISTER.append(cls.__name__)
    setattr(cls, "type", MATERIAL)
    if not hasattr(cls, "MID_FIELDNAME"):
        setattr(cls, "MID_FIELDNAME", "MID")
    if not hasattr(cls, "XID_FIELDNAME"):
        setattr(cls, "XID_FIELDNAME", getattr(cls, "MID_FIELDNAME"))
    return cls



import json
import logging
import shutil
import subprocess
from collections import defaultdict
from copy import deepcopy
from pathlib import Path

import yaml

import nastranio.cards as cards
from nastranio.constants import BULK, CASE, COMMENTS, EXEC, META, PARAMS, SUMMARY
from nastranio.readers.gmsh import GmsHParser, user_input
from nastranio.registry import Registry


def _convert_keys_to_int(d: dict):
    """recursively converts dictionnary keys to integer when possible.
    Usefull for json unserialization
    """
    new_dict = {}
    for k, v in d.items():
        try:
            new_key = int(k)
        except ValueError:
            new_key = k
        if type(v) == dict:
            v = _convert_keys_to_int(v)
        new_dict[new_key] = v
    return new_dict


class Study:
    def __init__(self, meshfile, autobuild=True):
        """
        Create a NASTRAN study based on GMSH mesh file.
        If autobuild is True (default), build the NASTRAN model.
        Setting `autobuild` to False allow to intercept GMSH
        mesh for modification (eg. nodes renumbering).
        """
        self._meshfile = Path(meshfile).expanduser()
        self.gmsh = GmsHParser(self._meshfile)
        logging.info("\n" + self.gmsh.info())
        if autobuild:
            self.build()

    def run(self, output="run", exist_ok=True):
        target = self._meshfile.parent / output
        target.mkdir(exist_ok=exist_ok)
        nasfile = target / (self._meshfile.stem + ".nas")
        self.to_nastran(nasfile)
        logging.info(f"wrote {nasfile}")
        MYSTRAN = shutil.which("mystran")
        try:
            subprocess.check_output([MYSTRAN, nasfile], stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError:
            logging.error("!!!")
        return target

    def load_user_params(self, params=None):
        """
        `params` can be:
          * a valid dictionnary
          * a path to JSON/YAML parameters file
          * a jsonified string
          * None (in which case, <meshfile>.params.json will be searched for)
        """
        self.ui = user_input.UserInput()
        if isinstance(params, dict):
            return self.load_user_params_from_dict(params)
        # None, string or Path. It may be either a JSON string, either a path
        # to parameters file.
        try:
            self.load_user_params_from_file(params)
        except (FileNotFoundError, OSError):
            self.load_user_params_from_json_txt(params)

    def load_user_params_from_dict(self, dic):
        return self.ui.load_parameters(dic.copy())

    def load_user_params_from_json_txt(self, json_txt, fmt, verbose=True):
        if fmt == ".json":
            params = json.loads(json_txt)
            params = _convert_keys_to_int(params)
        elif fmt in (".yaml", ".yml"):
            params = yaml.load(json_txt, Loader=yaml.FullLoader)
        if verbose:
            print(f"loading configuration from {fmt[1:]} string '{json_txt[:20]}...'")
        return self.load_user_params_from_dict(params)

    def _create_params_file(self, filepath):
        logging.warning(f"create {filepath} template file")
        params_data = self.get_ui_input_template()
        with open(filepath, "w") as fh:
            if filepath.suffix == ".json":
                fh.write(json.dumps(params_data, indent=4, sort_keys=False))
            if filepath.suffix in (".yaml", ".yml"):
                fh.write(yaml.dump(params_data, indent=4, sort_keys=False))

    def load_user_params_from_file(self, filepath, verbose=False):
        if filepath is None:
            extensions = (".json", ".yml", ".yaml")
            for suffix in extensions:
                filepath = self._meshfile.parent / (
                    self._meshfile.stem + ".params" + suffix
                )
                if filepath.exists():
                    break
            else:
                # no parameters file found
                filepath = self._meshfile.parent / (
                    self._meshfile.stem + ".params.json"
                )
                self._create_params_file(filepath)
        with open(filepath, "r") as fh:
            json_txt = fh.read()
        if verbose:
            print(f"loading configuration from {filepath}")
        return self.load_user_params_from_json_txt(
            json_txt, fmt=filepath.suffix, verbose=False
        )

    def orphan_nodes(self, clean=False):
        """delete orphan nodes from GmsH mesh"""
        used_gids = set(self.reg.mesh.gid2eids())
        defined_gids = set(self.gmsh.nodes)
        orphans = defined_gids - used_gids
        if orphans:
            logging.warning(f"{orphans=}")
            if clean:
                for orphan_gid in orphans:
                    self.gmsh.nodes.pop(orphan_gid)
        self.build()

        return orphans

    def get_ui_input_template(self):
        default = self.ui.get_default_template()
        # -------------------------------------------------------------
        # process boundaries. Everything found go to SID 1
        default["attributes"]["boundaries"][1] = {}
        _default_value = self.ui.get_default_boundary_input()
        master_key = "boundaries"
        find_in_grpname = "spc"
        self._update_default_master_key(
            master_key, find_in_grpname, _default_value, default
        )
        # -------------------------------------------------------------
        # process loading.  Everithing found go to SID 1
        default["attributes"]["loading"][1] = {}
        _default_value = self.ui.get_default_loading_input()
        master_key = "loading"
        find_in_grpname = "load"
        self._update_default_master_key(
            master_key, find_in_grpname, _default_value, default
        )
        # -------------------------------------------------------------
        # process affectations
        for dim, grps in self.gmsh.dim2physical_group.items():
            if dim == 0:
                continue
            for grp in grps:
                if dim == 1:
                    default["attributes"]["affectations"][grp.name] = {
                        "PID": 1,
                        "GO/X1": 1.0,
                        "CID": 0,
                        "X2": 0.0,
                        "X3": 0.0,
                    }
                elif dim == 2:
                    default["attributes"]["affectations"][grp.name] = {"PID": 3}
        return default

    def build(self, clean_orphan_nodes=False, bind_mesh_api=True):
        """
        mesh file (.msh)  ━───────┬───────>   NASTRAN BULK
        params file (.yml) ━──────┘
        """
        logging.info("building NASTRAN model")
        if not hasattr(self, "ui"):
            self.load_user_params()
        if clean_orphan_nodes:
            self.orphan_nodes(clean=True)

        self.reg = Registry()
        self.reg.container = {
            META.title: {"source": self._meshfile},
            BULK.title: {},
        }
        self._make_exec()
        self._make_cases()
        self._make_params()
        self._make_materials()
        self._make_properties()
        self._make_boundaries()
        self._make_loading()
        self._make_nodes()
        self._make_elements()
        # finalize registry
        self.reg.bind_mesh_api()
        self.reg._build_summary()

    def to_nastran(self, target=None):
        if not target:
            target = self._meshfile.parent / (self._meshfile.stem + ".nas")
        elif target == str:
            target = None
        return self.reg.to_nastran(target)

    def _update_default_master_key(self, master_key, find, default_value, default):
        all_groups = self.gmsh.physical_groups
        for (grpdim, grpid), grp in all_groups.items():
            if find in grp.name.lower():
                bound = {grp.name: default_value}
                default["attributes"][master_key][1].update(bound)
        # no group was named with SPC, pick-up a random one
        if not default["attributes"][master_key][1]:
            grp = next(iter(all_groups))  # pick up randomly
            bound = {grp.name: default_value}
            default["attributes"][master_key][1].update(bound)

    def _make_nodes(self):
        self.reg.container[BULK.title]["GRID"] = cards.GRID()
        for gid, grid in self.gmsh.nodes.items():
            self.reg.container["bulk"]["GRID"].append_sparams(
                {"ID": gid, "X1": grid.x, "X2": grid.y, "X3": grid.z}
            )

    def _finish_make_items(self, carddata):
        card = carddata["card"]
        cardname = card.__name__
        params = carddata["params"]
        over, grpname = carddata.get("over", (None, None))
        repeat = {}
        if over and (repeated := carddata.get("repeated")):
            if over == "nodes":
                ids = self.gmsh.physical_group2gids[grpname]
            else:
                ids = self.gmsh.physical_group2eids[grpname]
            repeat = {repeated[0]: ids}
        elif repeated := carddata.get("repeated"):
            # eg. PCOMP card has repeated fields without over
            repeat = repeated
        if cardname not in self.reg.container["bulk"]:
            self.reg.container["bulk"][cardname] = card()
        self.reg.bulk[cardname].append_sparams(params, **repeat)

    def _make_materials(self):
        # apply boundaries
        materials = self.ui.get_materials()
        for mid, material in materials.items():
            self._finish_make_items(material)

    def _make_properties(self):
        # apply boundaries
        properties = self.ui.get_properties()
        for pid, prop in properties.items():
            self._finish_make_items(prop)

    def _make_boundaries(self):
        # apply boundaries
        boundaries = self.ui.get_boundaries()
        for sid, constraints in boundaries.items():
            for constraint in constraints:
                self._finish_make_items(constraint)

    def _make_loading(self):
        self._loaded_gids = {}
        # apply boundaries
        loading = self.ui.get_loading()
        for sid, items in loading.items():
            sid2loaded_gids = defaultdict(list)
            self._loaded_gids[sid] = sid2loaded_gids
            for item in items:
                as_sum = item.get("unknown", {}).get("as_sum", False)
                if "over" in item:
                    gids = self.gmsh.physical_group2gids[item["over"][1]]
                    # only FORCE are allowed to be devided
                    if item["card"].__name__ == "FORCE" and as_sum:
                        item["params"]["N1"] /= len(gids)
                        item["params"]["N2"] /= len(gids)
                        item["params"]["N3"] /= len(gids)
                    for gid in gids:
                        _item = deepcopy(item)
                        _item["params"]["G"] = gid
                        _item["missing"].remove("G")
                        sid2loaded_gids[gid].append(_item)
                        # delegate _finish_make_item(_item) to deduplicator
                else:
                    # eg. GRAV
                    self._finish_make_items(item)
            self._loaded_gids[sid] = dict(sid2loaded_gids)
        self._make_loading_inspect_loaded_gids()

    def _make_loading_inspect_loaded_gids(self):
        duplicated_loading = defaultdict(dict)
        for sid, loaded_nodes in self._loaded_gids.items():
            for gid, items in loaded_nodes.items():
                if len(items) > 1:
                    duplicated_loading[sid][gid] = []
                    for item in items:
                        grpname = item["over"][1]
                        is_exclusive = item.get("unknown", {}).get("exclusive", False)
                        duplicated_loading[sid][gid].append((grpname, is_exclusive))
        duplicated_loading = dict(duplicated_loading)
        # ---------------------------------------------------------------------
        # process exclusivity
        if duplicated_loading:
            for sid, gids2loading_grp in list(duplicated_loading.items())[:]:
                for gid, grpnames in gids2loading_grp.copy().items():
                    dedup = defaultdict(set)
                    for grpname, is_exclusive in grpnames:
                        dedup[is_exclusive].add(grpname)
                    if True not in dedup:
                        continue
                    grpname_to_keep = dedup[True]
                    if len(grpname_to_keep) > 1:
                        raise ValueError(
                            f"several exclusivities detected for SID {sid} and GID {gid}: {grpname_to_keep}"
                        )
                    grpname_to_keep = next(iter(grpname_to_keep))
                    # remove unwanted items form _loaded_gids
                    to_pop = []
                    for unwanted_loading_grp in dedup[False]:
                        for i, item in enumerate(self._loaded_gids[sid][gid][:]):
                            item_grpname = item["over"][1]
                            if item_grpname == unwanted_loading_grp:
                                logging.info(
                                    f"loading SID{sid} GRID{gid}: ensure exclusivity for group '{grpname_to_keep}' by removing loading group '{item_grpname}'"
                                )
                                to_pop.append(i)
                    for i in to_pop[::-1]:
                        self._loaded_gids[sid][gid].pop(i)
                        duplicated_loading[sid][gid].pop(i)
                    if len(duplicated_loading[sid][gid]) == 1:
                        duplicated_loading[sid].pop(gid)
        # ---------------------------------------------------------------------
        # create loading
        for sid, loaded_nodes in self._loaded_gids.items():
            for gid, items in loaded_nodes.items():
                for item in items:
                    self._finish_make_items(item)
        # finally clean duplicated_loading
        duplicated_loading = {
            sid: dup for sid, dup in duplicated_loading.items() if dup
        }
        # ---------------------------------------------------------------------
        # feedback
        if duplicated_loading:
            logging.debug(f"{duplicated_loading=}")
            logging.warning("duplicated loading (combined) is detected")
            logging.info("you may want `exclusive` to avoid loads combination.")
            for sid, gids2loading_grp in duplicated_loading.items():
                msg = f"Duplicated loaded grids for SID {sid}:\n"
                for gid, grpnames in gids2loading_grp.items():
                    txt = " & ".join(["'%s'" % grp[0] for grp in grpnames])
                    msg += f"  * GRID {gid}: {txt}\n"
            logging.info(msg)
        self._dup_gids_loading = duplicated_loading

    def _make_elements(self):
        affect = self.ui.get_affectations()
        affected = set()
        for physicalName, item in affect.items():
            eids = self.gmsh.physical_group2eids[physicalName]
            for eid in eids:
                elt = self.gmsh.elements[eid]
                eid_item = deepcopy(item[elt.element_type])
                XID = eid_item["card"].XID_FIELDNAME
                eid_item["params"][XID] = eid
                eid_item["missing"].remove(XID)
                eid_item.pop("over")
                grids = dict(zip(eid_item["card"]._gids_header(), elt.gids))
                eid_item["params"].update(grids)
                eid_item.pop("missing")
                self._finish_make_items(eid_item)
                affected.add(eid)
        # give a feedback about physicalNames not set
        unset = {}
        for physicalName, eids in self.gmsh.physical_group2eids.items():
            nb_unset = len(eids - affected)
            if nb_unset:
                unset[physicalName] = (nb_unset, len(eids))
        for physicalName, (nb_unset, nb_total) in unset.items():
            msg = f"physical group '{physicalName}' has {nb_unset} non-affected elements out of {nb_total}"
            if "spc" not in physicalName.lower() and "load" not in physicalName.lower():
                logging.info(msg)
            else:
                logging.debug(msg)

    def _make_exec(self):
        self.reg.container["exec"] = self.ui.parameters["exec"]

    def _make_cases(self):
        cases = self.ui.parameters["cases"]
        boundary_sids = sorted(list(self.ui.get_boundaries().keys()))
        loading_sids = sorted(list(self.ui.get_loading().keys()))
        default_boundary_sid = boundary_sids.pop(0)
        default_loading_sid = loading_sids[0]
        if len(loading_sids) > 0:
            for sid in loading_sids:
                cases[f"SUBCASE {sid}"] = {
                    "SUBTITLE": f"LCID#{sid}",
                    "SPC": default_boundary_sid,
                    "LOAD": sid,
                }
        else:
            cases["default"]["SPC"] = default_boundary_sid
            cases["default"]["LOAD"] = default_loading_sid
        self.reg.container["cases"] = cases
        pass

    def _make_params(self):
        self.reg.container["params"] = self.ui.parameters["params"]
        pass



{
    "CBAR": {
        "EID": int,
        "G0": None,
        "G0/X1": float,
        "GA": int,
        "GB": int,
        "PA": None,
        "PB": None,
        "PID": int,
        "W1A": float,
        "W1B": float,
        "W2A": float,
        "W2B": float,
        "W3A": float,
        "W3B": float,
        "X1": float,
        "X2": float,
        "X3": float,
    },
    "CBUSH": {
        "CID": int,
        "EID": int,
        "GA": int,
        "GB": int,
        "GO/X1": float,
        "OCID": int,
        "PID": int,
        "S": float,
        "S1": None,
        "S2": None,
        "S3": None,
        "X2": float,
        "X3": float,
    },
    "CONM2": {
        "CID": int,
        "EID": int,
        "G": int,
        "I11": float,
        "I21": float,
        "I22": float,
        "I31": float,
        "I32": float,
        "I33": float,
        "M": float,
        "X1": float,
        "X2": float,
        "X3": float,
    },
    "CORD2C": {
        "A1": float,
        "A2": float,
        "A3": float,
        "B1": float,
        "B2": float,
        "B3": float,
        "C1": float,
        "C2": float,
        "C3": float,
        "CID": int,
        "RID": int,
    },
    "CORD2R": {
        "A1": float,
        "A2": float,
        "A3": float,
        "B1": float,
        "B2": float,
        "B3": float,
        "C1": float,
        "C2": float,
        "C3": float,
        "CID": int,
        "RID": int,
    },
    "CORD2S": {
        "A1": float,
        "A2": float,
        "A3": float,
        "B1": float,
        "B2": float,
        "B3": float,
        "C1": float,
        "C2": float,
        "C3": float,
        "CID": int,
        "RID": int,
    },
    "CQUAD4": {
        "EID": int,
        "G1": int,
        "G2": int,
        "G3": int,
        "G4": int,
        "MCID": int,
        "PID": int,
        "T1": None,
        "T2": None,
        "T3": None,
        "T4": None,
        "TFLAG": None,
        "THETA": float,
        "THETA/MCID": int,
        "ZOFFS": None,
    },
    "CROD": {"EID": int, "G1": int, "G2": int, "PID": int, "SM": None},
    "CTRIA3": {
        "EID": int,
        "G1": int,
        "G2": int,
        "G3": int,
        "MCID": int,
        "PID": int,
        "T1": None,
        "T2": None,
        "T3": None,
        "TFLAG": None,
        "THETA": float,
        "THETA/MCID": float,
        "ZOFFS": None,
    },
    "FORCE": {
        "CID": int,
        "F": float,
        "G": int,
        "N1": float,
        "N2": float,
        "N3": float,
        "SID": int,
    },
    "GRAV": {
        "A": float,
        "CID": int,
        "MB": int,
        "N1": float,
        "N2": float,
        "N3": float,
        "SID": int,
    },
    "GRID": {
        "CD": int,
        "CP": int,
        "ID": int,
        "PS": int,
        "SEID": None,
        "X1": float,
        "X2": float,
        "X3": float,
    },
    "LOAD": {"S": float, "SID": int, "load_factorsID": int},
    "MAT1": {
        "A": float,
        "E": float,
        "G": float,
        "GE": float,
        "MCSID": None,
        "MID": int,
        "NU": float,
        "RHO": float,
        "SC": None,
        "SS": None,
        "ST": None,
        "TREF": float,
    },
    "MAT8": {
        "A1": float,
        "A2": float,
        "E1": float,
        "E2": float,
        "F12": None,
        "G12": float,
        "G1Z": float,
        "G2Z": float,
        "GE": None,
        "MID": int,
        "NU12": float,
        "RHO": float,
        "S": None,
        "STRN": None,
        "TREF": float,
        "Xc": None,
        "Xt": None,
        "Yc": None,
        "Yt": None,
    },
    "PBAR": {
        "A": float,
        "C1": float,
        "C2": float,
        "D1": float,
        "D2": float,
        "E1": float,
        "E2": float,
        "F1": float,
        "F2": float,
        "I1": float,
        "I12": float,
        "I2": float,
        "J": float,
        "K1": float,
        "K2": float,
        "MID": int,
        "NSM": float,
        "PID": int,
    },
    "PBUSH": {
        '"B"': str,
        '"GE"': str,
        '"K"': str,
        '"RCV"': str,
        "B1": None,
        "B2": None,
        "B3": None,
        "B4": None,
        "B5": None,
        "B6": None,
        "EA": None,
        "ET": None,
        "GE1": None,
        "GE2": None,
        "GE3": None,
        "GE4": None,
        "GE5": None,
        "GE6": None,
        "K1": float,
        "K2": float,
        "K3": float,
        "K4": float,
        "K5": float,
        "K6": float,
        "PID": int,
        "SA": None,
        "ST": None,
    },
    "PCOMP": {
        "FT": None,
        "GE": float,
        "LAM": None,
        "NSM": float,
        "PID": int,
        "SB": None,
        "TREF": None,
        "Z0": None,
        "pcomp_layupID": int,
    },
    "PLOAD4": {
        "CID": None,
        "EID": int,
        "G1": None,
        "G1orG4": None,
        "N1": None,
        "N2": None,
        "N3": None,
        "P1": float,
        "P2": None,
        "P3": None,
        "P4": None,
        "SID": int,
    },
    "PROD": {"A": float, "C": float, "J": float, "MID": int, "NSM": float, "PID": int},
    "PSHELL": {
        "12I/T**3": float,
        "MID1": int,
        "MID2": int,
        "MID3": int,
        "MID4": None,
        "NSM": float,
        "PID": int,
        "T": float,
        "TS/T": float,
        "Z1": None,
        "Z2": None,
    },
    "RBE2": {"ALPHA": float, "CM": int, "EID": int, "GN": int, "rbe2_gidsetID": int},
    "RBE3": {
        "ALPHA": float,
        "EID": int,
        "REFC": int,
        "REFGRID": int,
        "rbe3_umID": int,
        "rbe3_wcgID": int,
    },
    "SPC1": {"C": int, "SID": int, "spc1_gridsetID": int},
}



"""
interface with OP2 results
"""

# ============================================================================
# ⚠ NOTE ⚠
# ----------------------------------------------------------------------------
# For now, use pyNastran
# ----------------------------------------------------------------------------
# nic@alcazar -- lundi 2 décembre 2019, 14:08:12 (UTC+0100)
# mercurial: 36fcd4a16748 tip
# ============================================================================

import glob
import logging
import multiprocessing as mp
import os
import time

import numpy as np
import pandas as pd

try:
    from pyNastran.op2.op2 import OP2
except ImportError as exc:
    logging.warning("pyNastran not installed.")
    import sys

    sys.exit(0)  # make pytest fails..

PD_OPTIONS = {"precision": 3}
NP_PRINTOPTIONS = {"precision": 3, "threshold": 20}


def set_options(pd_options=None, np_printoptions=None):
    if not pd_options:
        pd_options = PD_OPTIONS
    for k, v in pd_options.items():
        pd.set_option(k, v)
    if not np_printoptions:
        np_printoptions = NP_PRINTOPTIONS
    for k, v in np_printoptions.items():
        np.set_printoptions(**np_printoptions)


class MultipleOP2:
    """basic wrapper around pyNastran OP2 cls"""

    def __init__(self, mode="nx", autorename=True):
        self._mode = mode
        self._autorename = autorename
        self.subcases = {}
        # automatic renaming for DataFrames columns headers
        self._op2nasca = {
            "lc": "lc",
            "NodeID": "gid",
            "ElementID": "eid",
            "ElementType": "source",
            "f1": "t1",
            "f2": "t2",
            "f3": "t3",
            "m1": "r1",
            "m2": "r2",
            "m3": "r3",
            "t12": "shear_xy_mat",
            "t1z": "shear_xz_mat",
            "t2z": "shear_yz_mat",
            "NodeID": "gid",
            "ElementID": "eid",
            "Item": "vector",
            "Layer": "ply_id",
            0: "FEA_value",
        }
        self._nasca2op = {v: k for k, v in self._op2nasca.items()}

    def _new_op2(self, debug=False, log=None, debug_file=None):
        """op2 boilerplate"""
        op2 = OP2(debug=debug, log=log, debug_file=debug_file)
        op2.set_mode(self._mode)
        return op2

    @property
    def lcs(self):
        return self.subcases.copy()

    def read_op2(self, args):
        """read a single op2. Can be called multiple times"""
        if isinstance(args, str):
            multiprocess = False
            filename = args
        else:
            multiprocess = True
            filename = args[0]
            container = args[1]
        fname, ext = os.path.splitext(filename)
        rootdir, _ = os.path.split(filename)
        op2 = self._new_op2()
        op2.read_op2(
            filename, combine=True, build_dataframe=False, skip_undefined_matrices=True
        )
        if not multiprocess:
            # single process reading
            # get last id from subcases
            try:
                id_offset = max(self.subcases.keys())
            except:
                id_offset = 0
            for isubcase in op2.subcase_key.keys():
                self.subcases[isubcase + id_offset] = (op2, isubcase, op2.title)
        else:
            container.append((filename, op2, tuple(op2.subcase_key.keys())))

    def read_op2_in_dir(self, path, pattern="*.op2", multiprocess=False):
        """
        Read several op2 from dir, as per provided pattern. Default (``'*.op2'``) is to
        read all op2 from path."""
        files = glob.glob(os.path.join(path, pattern))
        files = sorted(files)
        start = time.time()
        if not multiprocess:
            for file in files:
                self.read_op2(file)
            stop = time.time()
        else:
            # multiprocess reading
            with mp.Manager() as manager:
                op2s = manager.list()
                args = [(file, op2s) for file in files]
                with manager.Pool() as pool:
                    pool.map(self.read_op2, args)
                op2s = list(op2s)
            # sort op2s by filename
            op2s = sorted(op2s, key=lambda x: x[0])
            offset = 0
            for (fname, op2, lcids) in op2s:
                for isubcase in lcids:
                    self.subcases[isubcase + offset] = (op2, isubcase, op2.title)
                offset += len(lcids)
            stop = time.time()
        logging.info(70 * "-")
        logging.info("read %d files in %.3f sec." % (len(files), stop - start))
        logging.info(70 * "-")

    def release_df(self, df):
        """post-processing function for released DataFrames"""
        if self._autorename:
            df = df.reset_index().rename(columns=self._op2nasca)
            cols = [c for c in df.columns if not c.startswith("level_")]
            df = df[cols]
            df.columns.name = None
            return df
        return df.reset_index()

    def result(self, attr, filter_v=None, autoclean=True):
        """read one single result from all op2s and all subcases,
        and return a dataframe
        """
        _res = {}
        for lcid, (op2, isubcase, title) in self.subcases.items():
            logging.info(
                f'recover "{attr}" for lcid {lcid} "{title}" (isubcase={isubcase})',
                end="...",
            )
            res = getattr(op2, attr)
            res = res.get(isubcase)
            if not res:
                logging.info("no results. Return")
                return
            if res.dataframe is None:
                res.build_dataframe()
            res = res.dataframe.copy().reset_index()
            if filter_v:
                res = res[res[list(filter_v.keys())].isin(filter_v).all(axis=1)]
            _res[lcid] = res
            logging.info("ok")
        _res = pd.concat(_res, axis=0)
        _res.index.names = ["lc"] + _res.index.names[1:]
        if autoclean:
            return self.release_df(_res)
        return _res

    def results(self, attrs, filter_v=None):
        """sequential `self.result()`` calls"""
        dfs = [self.result(attr, filter_v=filter_v, autoclean=False) for attr in attrs]
        # eventually remove Nones
        dfs = [df for df in dfs if df is not None]
        df = pd.concat(dfs)
        return self.release_df(df)

    def gpf(self, filter_v=None):
        """
        read 'grid_point_forces'
        """
        filter_v = self._translate_filter(filter_v)
        return self.result("grid_point_forces", filter_v=filter_v)

    def composites(self, filter_v=None):
        """"""
        attrs = (
            "cquad4_composite_stress",
            "ctria3_composite_stress",
            "cquad8_composite_stress",
        )
        filter_v = self._translate_filter(filter_v)
        df = self.results(attrs, filter_v=filter_v)
        return df

    def _translate_filter(self, filter_v):
        if not filter_v:
            return
        translated = {}
        for k, values in filter_v.items():
            values = [self._nasca2op.get(v, v) for v in values]
            translated[self._nasca2op.get(k, k)] = values
        return translated



"""
NASTRAN loading cards collection.

This include loading and constraints, therfore, GRIDS cards
"""

import logging
import re
from collections import defaultdict
from pprint import pprint

import numpy as np
import pandas as pd
from numtools.vgextended import loc_array

from nastranio.cardslib import SimpleCard, SimpleCyclingCard
from nastranio.decorators import boundary, cached_property, loading_type


@loading_type()
class GRAV(SimpleCard):
    """
    Acceleration or Gravity Load
    Defines acceleration vectors for gravity or other acceleration loading.

    ref: NX Nastran 12 Quick Reference Guide 14-55 (p.1921)

    >>> grav = GRAV()
    >>> bulk = ["GRAV           1       0      1.     -9.      0.      0.",
    ...         "GRAV           1       0      1.     -9.      0.      0.       5"]
    >>> for b in bulk: grav.parse(b)
    >>> pprint(grav.export_data()['main'])
    {'A': [1.0, 1.0],
     'CID': [0, 0],
     'MB': [0, 5],
     'N1': [-9.0, -9.0],
     'N2': [0.0, 0.0],
     'N3': [0.0, 0.0],
     'SID': [1, 1]}
    """

    TABLE = """
| 1    | 2   | 3   | 4 | 5  | 6  | 7  | 8  | 9 | 10 |
|------+-----+-----+---+----+----+----+----+---+----|
| GRAV | SID | CID | A | N1 | N2 | N3 | MB |   |    |
    """
    DEFAULTS = {"CID": 0, "MB": 0}


@loading_type("nodal")
class FORCE(SimpleCard):
    """
    Static Force
    Defines a static concentrated force at a grid point by specifying a vector.

    ref: NX Nastran 12 Quick Reference Guide 14-10 (p.1876)
    """

    TABLE = """
    | 1     | 2   | 3 | 4   | 5 | 6  | 7  | 8  | 9 | 10 |
    |-------+-----+---+-----+---+----+----+----+---+----|
    | FORCE | SID | G | CID | F | N1 | N2 | N3 |   |    |
    """
    DEFAULTS = {"cid": 0, "N1": 0.0, "N2": 0.0, "N3": 0.0}


@loading_type("nodal")
class MOMENT(SimpleCard):
    """
    Static Moment
    Defines a static concentrated moemnt at a grid point by specifying a vector.

    ref: NX Nastran 12 Quick Reference Guide 15-127 (p.2093)
    """

    TABLE = """
    | 1      | 2   | 3 | 4   | 5 | 6  | 7  | 8  | 9 | 10 |
    |--------+-----+---+-----+---+----+----+----+---+----|
    | MOMENT | SID | G | CID | M | N1 | N2 | N3 |   |    |
    """
    DEFAULTS = {"cid": 0, "N1": 0.0, "N2": 0.0, "N3": 0.0}


@loading_type()
class LOAD(SimpleCyclingCard):
    """
    Static Force
    Defines a static concentrated force at a grid point by specifying a vector.

    ref: NX Nastran 12 Quick Reference Guide 14-90 (p.1956)
    """

    TABLE = """
    | 1    | 2   | 3  | 4      | 5  | 6  | 7  | 8  | 9  | 10 |
    |------+-----+----+--------+----+----+----+----+----+----|
    | LOAD | SID | S  | S1     | L1 | S2 | L2 | S3 | L3 |    |
    |      | S4  | L4 | -etc.- |    |    |    |    |    |    |
    """
    REPEATED_DATA_NAME = "FACTORS"


@loading_type("elemental")
class PLOAD4(SimpleCard):
    """
    Pressure Load on Shell and Solid Element Faces
    Defines a pressure load on a face of a CHEXA, CPENTA, CTETRA, CPYRAM,
    CTRIA3, CTRIA6, CTRIAR, CQUAD4, CQUAD8, or CQUADR element.

    ref: NX Nastran 12 Quick Reference Guide 16-164 (p.2394)

    Alternate format (using "THRU") not taken into account.
    """

    TABLE = """
    | 1      | 2   | 3   | 4  | 5  | 6  | 7  | 8  | 9      | 10 |
    |--------+-----+-----+----+----+----+----+----+--------+----|
    | PLOAD4 | SID | EID | P1 | P2 | P3 | P4 | G1 | G3orG4 |    |
    |        | CID | N1  | N2 | N3 |    |    |    |        |    |
    """


@boundary
class SPC1(SimpleCyclingCard):
    """
    Single-Point Constraint, Alternate Form
    Defines a set of single-point constraints.

    ref: NX Nastran 12 Quick Reference Guide 17-174 (p.2648)

    >>> spc1 = SPC1()
    >>> bulk = ["SPC1         290     123   12508",
    ...         "SPC1         292     123      12",
    ...         "SPC1         291  123456   12509       2       3       4       5"]
    >>> for b in bulk: spc1.parse(b)
    >>> pprint(spc1.export_data()['main'])  # doctest: +NORMALIZE_WHITESPACE
    {'C': [123, 123, 123456], 'SID': [290, 292, 291], 'spc1_gridsetID': [0, 1, 2]}
    >>> pprint(spc1.export_data()['spc1_gridset'])  # doctest: +NORMALIZE_WHITESPACE
    [[{'G': 12508}],
     [{'G': 12}],
     [{'G': 12509}, {'G': 2}, {'G': 3}, {'G': 4}, {'G': 5}]]
    """

    REPEATED_DATA_NAME = "GRIDSET"
    TABLE = """
    | 1    | 2   | 3  | 4  | 5      | 6  | 7  | 8  | 9  | 10 |
    |------+-----+----+----+--------+----+----+----+----+----|
    | SPC1 | SID | C  | G1 | G2     | G3 | G4 | G5 | G6 |    |
    |      | G7  | G8 | G9 | -etc.- |    |    |    |    |    |
    """


@boundary
class GRID(SimpleCard):
    """
    Grid Point
    Defines the location of a geometric grid

    ref: NX Nastran 12 Quick Reference Guide 14-60 (p.1926)
    """

    TABLE = """
        | 1    | 2  | 3  | 4  | 5  | 6  | 7  | 8  | 9    | 10 |
        |------+----+----+----+----+----+----+----+------+----|
        | GRID | ID | CP | X1 | X2 | X3 | CD | PS | SEID |    |
    """
    DEFAULTS = {
        "CP": 0,
        "CD": 0,
        "PS": None,
        "SEID": None,
        "X1": 0.0,
        "X2": 0.0,
        "X3": 0.0,
    }
    XID_FIELDNAME = "ID"

    def to_vtk(self, **kwargs):
        """return grids as VTK legacy format"""
        from io import StringIO

        fh = StringIO()
        msg = ["POINTS %d double" % len(self)]
        coords = self.coords()[1]
        np.savetxt(fh, coords, **kwargs)
        fh.seek(0)
        msg += fh.readlines()
        return msg

    def coords(self, incsys=None, csysreg=None, gids=None, asdf=False):
        """
        return two numpy arrays: one vector of eids and one 3xN matrix of coordinates
        """
        cdefined = self.array[["ID", "CP", "X1", "X2", "X3"]]  # coordinates as defined
        _gids = cdefined["ID"]
        xyz = np.vstack((cdefined["X1"], cdefined["X2"], cdefined["X3"])).T
        csys = cdefined["CP"]
        if gids:
            if not isinstance(gids, (list, np.ndarray)):
                gids = list(gids)
        # ====================================================================
        # calculate coordinates in requested CSYS
        # ====================================================================
        if incsys:  # not None and not 0
            if incsys == -1:  # way for _translate_grids_to_0 to bypass
                incsys = 0
            _initial_gids = _gids.copy()
            # which points are not concerned?
            nop = csys == incsys
            nop_gids, _gids = _gids[nop], _gids[~nop]
            nop_xyz, xyz = xyz[nop], xyz[~nop]
            nop_csys, csys = csys[nop], csys[~nop]
            origins = set(csys)  # this will define how many change we will perform
            for origin_csys in origins:
                csys_obj = csysreg[origin_csys]
                # search all the points defined in origin_csys
                which = csys == origin_csys
                c_gids, c_xyz, c_csys = _gids[which], xyz[which], csys[which]
                c_xyz = csys_obj.export_vertices(c_xyz, to_csys=incsys)
                c_csys.fill(incsys)
                # nothing to change to c_gids
                nop_xyz = np.vstack((nop_xyz, c_xyz))
                nop_gids = np.hstack((nop_gids, c_gids))
                nop_csys = np.hstack((nop_csys, c_csys))
            xyz, _gids, csys = nop_xyz, nop_gids, nop_csys
            # ----------------------------------------------------------------
            # sort arrays as initially sorted
            positions = loc_array(_initial_gids, _gids)
            np.put(_gids, positions, _gids)
            np.put(xyz[:, 0], positions, xyz[:, 0])
            np.put(xyz[:, 1], positions, xyz[:, 1])
            np.put(xyz[:, 2], positions, xyz[:, 2])
            np.put(csys, positions, csys)

        if gids:
            mask = np.isin(_gids, gids, assume_unique=True)
            _gids = _gids[mask]
            xyz = xyz[mask]
            csys = csys[mask]

        if len(set(csys)) > 1:
            logging.warning("return coords in different systems")

        if not asdf:
            return _gids, xyz, csys
        else:
            data = np.insert(xyz, 0, csys, axis=1)
            df = pd.DataFrame(data, columns=["csys", "X", "Y", "Z"], index=_gids)
            df.csys = df.csys.astype(int)
            df.index.names = ["gid"]
            return df


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



"""Generic helpers module for developers"""
import datetime as dt
import logging
import os
import time
from collections import defaultdict
from itertools import chain

import numpy as np

try:
    from deepdiff import DeepDiff

    ISDEEPDIFF = True
except ImportError:
    ISDEEPDIFF = False


def project_point_fast(point, ga, gb):
    """
    return a tuple (<projection point>, <`t`>)

    >>> ga = np.array([1.5, 5.5, 0])
    >>> gb = np.array([6, -1.5, 0])
    >>> proj, t= project_point_fast(np.array([9, -1, 0]), ga, gb)

    `proj` is the calcualted projected point:
    >>> proj
    array([ 6.64981949, -2.51083032,  0.        ])

    `t` is the gagb range factor:
        * t<0: projection is 'before' ga
        * t>1, projection is 'after' gb.
    >>> t
    1.144404332129964
    """
    gagb_vector = gb - ga
    t = np.sum((point - ga) * gagb_vector) / np.sum(gagb_vector**2)
    projection = ga + t * gagb_vector
    return projection, t


def project_point(point, ga, gb, strategy="on_extend", return_t=False):
    r"""
    project a point onto a segment [sega, segb] with different strategies:

      * 'strict': calculate projection only if projected point lies between
        ga and gb.
      * 'on_extend': if projection is "outside" of the segment,
        calculate its position on the "extended" segment.
      * 'on_end': if projection is "outside" of the segment,
        calculate the projection as beeing the closest extremity (ga or gb)
        of the theoretical projection.

       + ga
        \
         \
          \   + point
           +
            \
             + gb

    All of those strategies leads to the same result if the projection is
    between [ga, gb]

    >>> ga = np.array([1.5, 5.5, 0])
    >>> gb = np.array([6, -1.5, 0])


    For a point whose projection lies on ga, gb segment, strategy is skipped:

    >>> project_point(np.array([2, 1.5, 0]), ga, gb)
    array([3.46570397, 2.44223827, 0.        ])


    >>> project_point(np.array([9, -1, 0]), ga, gb, strategy='strict')
    Traceback (most recent call last):
        ...
    ValueError: point [9, -1, 0] does not project onto ga-gb line segment

    >>> project_point(np.array([9, -1, 0]), ga, gb, strategy='on_extend')
    array([ 6.64981949, -2.51083032,  0.        ])

    >>> project_point(np.array([9, -1, 0]), ga, gb, strategy='on_end')
    array([ 6. , -1.5,  0. ])
    """
    # distance between ga and gb
    projection, t = project_point_fast(point, ga, gb)
    has_changed = False
    if strategy != "on_extend" and (t > 1 or t < 0):
        if strategy == "strict":
            raise ValueError(
                f"point {point.tolist()} does not project onto ga-gb line segment"
            )
        elif strategy == "on_end":
            # if you need the point to project on line segment between ga and gb or closest point of the line segment
            has_changed = True
            t = max(0, min(1, t))
    if has_changed:
        gagb_vector = gb - ga
        projection = ga + t * gagb_vector
    # distance = np.sum((projection - point) ** 2)
    if not return_t:
        return projection
    return projection, t


def calcdiff(args, exclude_meta=True, **kwargs):
    d1, d2 = args
    if not ISDEEPDIFF:
        msg = "optional DeepDiff lib is not installed"
        raise RuntimeError(msg)
    else:
        if exclude_meta:
            exclude_paths = ["root['meta']", "root['meta']"]
        else:
            exclude_paths = None
        diff = DeepDiff(
            d1,
            d2,
            exclude_regex_paths=[
                r"root\['_.*'\]",
                r"root\[.*'\]\['_.*'\]",
                r"_cached.*",
            ],
            exclude_paths=exclude_paths,
            ignore_type_in_groups=((dict, defaultdict), (list, tuple)),
            ignore_nan_inequality=True,
            **kwargs,
        )
    return diff


class Chronos:
    """
    Easily sequence timing in yout function:

    >>> c = Chronos()
    >>> c.click('hello world')
    hello world: 0:00:00.0...
    >>> time.sleep(0.05)
    >>> c.click('another event')
    another event: 0:00:00.05...
    """

    def __init__(self):
        self._starts = dt.datetime.now()
        self._events = [{"abstime": self._starts, "reltime": 0, "event": "starts"}]

    def click(self, event):
        now = dt.datetime.now()
        reltime = now - self._events[0]["abstime"]
        self._events.append({"abstime": now, "reltime": reltime, "event": event})
        print("%s: %s" % (event, reltime))


# ============================================================================
# numpy structured arrays
# ============================================================================
def dic2array(data):
    """
    Convert a dictionnary to numpy structured array.

    >>> data ={'x': [1.5, 3., 1., 9.7],
    ...        'y': [2.5, 4, None, 5.2],
    ...        'z': [1, 1, 0, 0],
    ...        'w': ['A', 'd', 'B', None],
    ...        'v': [1, 2, None, None],
    ...        'u': [None, None, None, None]
    ...  }
    >>> dic2array(data)
    array([(1.5, 2.5, 1, 'A',  1., nan), (3. , 4. , 1, 'd',  2., nan),
       (1. , nan, 0, 'B', nan, nan), (9.7, 5.2, 0, 'N', nan, nan)],
      dtype=[('x', '<f8'), ('y', '<f8'), ('z', '<i8'), ('w', '<U1'), ('v', '<f8'), ('u', '<f8')])
    >>> import pandas as pd
    >>> pd.DataFrame(dic2array(data))
             x    y  z  w    v   u
    0  1.5  2.5  1  A  1.0 NaN
    1  3.0  4.0  1  d  2.0 NaN
    2  1.0  NaN  0  B  NaN NaN
    3  9.7  5.2  0  N  NaN NaN
    >>> pd.DataFrame(dic2array(data)).dtypes
    x    float64
    y    float64
    z      int64
    w     object
    v    float64
    u    float64
    dtype: object
    """
    values = []
    names = []
    formats = []
    type_None = type(None)
    for k, varray in data.items():
        types = {type(i) for i in varray}
        if str in types:
            formats.append("<U1")
        elif type_None in types or float in types or np.float64 in types:
            formats.append("<f8")
        elif int in types or np.int64 in types:
            formats.append("<i8")
        else:
            breakpoint()
            formats.append("<U1")
        values.append(tuple(varray))
        names.append(k)
    return np.array(list(zip(*values)), dtype={"names": names, "formats": formats})


def dic2array_legacy(data, nb_entries=None, None2NaN=True):
    """
    Convert a dictionnary to numpy structured array.

    >>> data ={'x': [1.5, 3., 1., 9.7],
    ...        'y': [2.5, 4, None, 5.2],
    ...        'z': [1, 1, 0, 0]}
    >>> dic2array(data)
    array([(1.5, 2.5, 1), (3. , 4. , 1), (1. , nan, 0), (9.7, 5.2, 0)],
          dtype=[('x', '<f8'), ('y', '<f8'), ('z', '<i8')])
    >>> data  # data has not been modified
    {'x': [1.5, 3.0, 1.0, 9.7], 'y': [2.5, 4, None, 5.2], 'z': [1, 1, 0, 0]}
    """
    data = {k: v.copy() for k, v in data.items()}
    if not nb_entries:
        krandom = list(data.keys())[0]  # pick-up a random key
        nb_entries = len(data[krandom])
    values = []
    for ix in range(nb_entries):
        col = []
        for fieldname in data.keys():
            val = data[fieldname][ix]
            if val is None and None2NaN:
                val = data[fieldname][ix] = np.NaN
            col.append(val)
        values.append(tuple(col))
    _data = {k: np.array(v) for k, v in data.items()}
    formats = [a.dtype.str for a in _data.values()]
    names = [k for k in data.keys()]

    array = np.array(values, dtype={"names": names, "formats": formats})
    return array


def array2dic(array, astype=None):
    """convert a structured array to regular dictionnary.
    If `astype` is not provided, returned values remain numpy 1D arrays
    >>> data ={'x': [1.5, 3., 1., 9.7],
    ...        'y': [2.5, 4, 3, 5.2],
    ...        'z': [1, 1, 0, 0]}
    >>> array2dic(dic2array(data), astype=list)
    {'x': [1.5, 3.0, 1.0, 9.7], 'y': [2.5, 4.0, 3.0, 5.2], 'z': [1, 1, 0, 0]}
    """
    ret = {}
    if not astype:
        for col in array.dtype.names:
            ret[col] = array[col]
    else:
        for col in array.dtype.names:
            ret[col] = astype(array[col])
    return ret


def bunch_legacy(dic):
    """given a dict {k: iterable}, return a set of all iterables

    >>> bunch({'a': {4, 5, 6}, 'b': set((1, 4, 7))})
    frozenset({1, 4, 5, 6, 7})
    """
    ret = set()
    for s in dic.values():
        if not hasattr(s, "__iter__"):
            s = set((s,))
        ret |= set(s)
    return frozenset(ret)


def bunch(dic):
    """given a dict {k: iterable}, return a set of all iterables

    >>> bunch({'a': {4, 5, 6}, 'b': set((1, 4, 7))})
    frozenset({1, 4, 5, 6, 7})
    """
    return frozenset(chain.from_iterable(dic.values()))


def object_attributes(obj, mode, blacklist=None):
    """list object attributes of a given type"""
    if not blacklist:
        blacklist = ()
    attrs = defaultdict(set)
    for k, v in dir(obj):
        if k in blacklist:
            continue
        if k.startswith("__"):
            attrs["protected"].add(k)
        elif k.startswith("_"):
            attrs["private"].add(k)
        else:
            attrs["public"].add(k)
    # ------------------------------------------------------------------------
    if mode in attrs:
        # public, private, protected
        return sorted(list(attrs[mode]))
    if mode == "both":
        return sorted(list(attrs["public"] | attrs["private"]))
    if mode == "all":
        return sorted(list(attrs["public"] | attrs["private"] | attrs["protected"]))
    raise KeyError(
        f'mode {mode} shall be one of {"public", "private", "protected", "both", "all"}'
    )


def check_path(filename, name="file") -> None:
    """checks that the file exists"""
    try:
        exists = os.path.exists(filename)
    except TypeError:
        msg = "cannot find %s=%r\n" % (name, filename)
        raise TypeError(msg)
    if not exists:
        msg = "cannot find %s=%r\n%s" % (name, filename)
        raise FileNotFoundError(msg)


def transform_dict_of_list(data):
    """transform a dict of list into a list of dict:

    >>> data = {'MIDi': [1002, 1003, 1009],
    ... 'SOUTi': ['NO', 'YES', 'YES'],
    ... 'Ti': [0.018, 0.339, 0.018],
    ... 'THETAi': [0.0, 0.0, 5.0]}
    >>> expected = [{'MIDi': 1002, 'SOUTi': 'NO', 'Ti': 0.018, 'THETAi': 0.0},
    ...             {'MIDi': 1003, 'SOUTi': 'YES', 'Ti': 0.339, 'THETAi': 0.0},
    ...             {'MIDi': 1009, 'SOUTi': 'YES', 'Ti': 0.018, 'THETAi': 5.0}]
    >>> transform_dict_of_list(data) == expected
    True
    """
    ret = []
    nb_items = len(data[next(iter(data.keys()))])
    for item_no in range(nb_items):
        _d = {}
        for fieldname, seq in data.items():
            _d[fieldname] = list(seq)[item_no]
        ret.append(_d)
    return ret


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from collections import defaultdict, namedtuple

# ----------------------------------------------------------------------------
# key names for container dict
_SECTION = namedtuple("SECTION", ["title", "type", "args"])
BULK = _SECTION("bulk", defaultdict, (list,))
EXEC = _SECTION("exec", dict, ())
CASE = _SECTION("cases", dict, ())
PARAMS = _SECTION("params", dict, ())
COMMENTS = _SECTION("comments", list, ())
META = _SECTION("meta", dict, ())
SUMMARY = _SECTION("summary", defaultdict, (set,))

# # cards families
# NODE = 'node'
ELEMENT = "element"
PROPERTY = "property"
MATERIAL = "material"
LOADING = "load"
BOUNDARY = "boundary"
AXIS = "axis"
UNKNOWN = "unknown cards"

# ----------------------------------------------------------------------------
# elements shapes: used by elements decorators
Shapes = namedtuple("SHAPE", ("VERTICE", "LINE", "TRIA", "QUAD", "MPC"))
shapes = Shapes(VERTICE="point", LINE="line", TRIA="triangle", QUAD="quad", MPC="mpc")

# ----------------------------------------------------------------------------
# VTK shapes
VTKShapes = {
    shapes.TRIA: "VTK_TRIANGLE",
    shapes.QUAD: "VTK_QUAD",
    shapes.LINE: "VTK_LINE",
    shapes.MPC: "VTK_LINE",
    shapes.VERTICE: "VTK_VERTEX",
}

# -----------------------------------------------------------------------------
# GMSH elemets type
GMSHElementTypes = {
    shapes.LINE: 1,
    shapes.TRIA: 2,
    shapes.QUAD: 3,
    shapes.VERTICE: 15,
}

# name (eg. "MAT1") of available cards. filled by cards decorator
CARDS_REGISTER = []



"""
UserInput Class provides a way to complete a gmsh mesh with mesh attributes

)


"""
from .ui_versions.user_input_v2 import UserInput

if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE)



import logging
from functools import wraps

import numpy as np
from rtree import index

RTREE_COORDS_DEC = 2


def check_coords(*argnames):
    def _check_coords(fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            for argname in argnames:
                kwargs[argname] = list(
                    np.round(kwargs[argname], decimals=RTREE_COORDS_DEC)
                )
            return fn(*args, **kwargs)

        return wrapper

    return _check_coords


def list_content(rtree):
    for item in list(rtree.intersection(rtree.bounds, objects=True)):
        print(item)


class RTrees:
    """
    https://rtree.readthedocs.io/en/latest/tutorial.html
    """

    def __init__(self, registry, index_elems=("0d", "1d", "2d", "3d")):
        p = index.Property()
        p.dimension = 3
        self._rtree_properties = p
        # p.dat_extension = 'data'
        # p.idx_extension = 'index'
        self._ixs = {"GRID": index.Index(properties=self._rtree_properties)}
        self.index_elems = index_elems
        self.reg = registry

    def list(self, cardname=None, rtree=None):
        if rtree is None:
            rtree = self._ixs[cardname]
        return tuple(rtree.intersection(rtree.bounds, objects=True))

    def create_rtree(self, eids=None, gids=None):
        """create and return 'on-the-fly' a new rtree"""
        if eids is not None and gids is not None:
            raise ValueError("both `eids` and `gids` cannot be set")
        _rtree = index.Index(properties=self._rtree_properties)
        ids = eids if eids is not None else gids
        if gids is not None:
            src_rtrees = (self._ixs["GRID"],)
        else:
            src_rtrees = []
            for cardname, rtree in self._ixs.items():
                if cardname == "GRID":
                    continue
                src_rtrees.append(rtree)
        for rtree in src_rtrees:
            for item in list(rtree.intersection(rtree.bounds, objects=True)):
                if item.id in ids:
                    _rtree.insert(item.id, item.bbox)
        return _rtree

    # check_coords("coords")
    def add_grid(self, gid, /, coords, rtree=None):
        """insert gid in RTREE index"""
        self._ixs["GRID"].insert(int(gid), coords)
        if rtree:
            rtree.insert(int(gid), coords)

    # check_coords("coords")
    def add_elem(self, cardname, eid, /, coords, rtree=None):
        try:
            # if a pandas.Series is passed
            coords = coords.values
        except AttributeError:
            pass
        self._ixs[cardname].insert(int(eid), coords)
        if rtree:
            rtree.insert(int(eid), coords)

    # check_coords("old_coords", "coords")
    def _update(self, cardname, id, /, old_coords, coords, rtree=None):
        rt = self._ixs.get(cardname)
        if rt is None and rtree is None:
            return
        rt.delete(id, old_coords)
        rt.insert(id, coords)
        if rtree:
            rtree.delete(id, old_coords)
            rtree.insert(id, coords)

    def get_bbox(self, cardname, id):
        ix = self._ixs[cardname]
        for f in ix.intersection(ix.bounds, objects=True):
            if f.id == id:
                return f.bbox

    def get_gid_bbox(self, gid):
        return self.get_bbox(cardname="GRID", id=gid)
        ix = self._ixs["GRID"]
        for f in ix.intersection(ix.bounds, objects=True):
            if f.id == id:
                return f.bbox

    def get_eid_bbox(self, eid, cardname=None):
        if cardname:
            return self.get_bbox(cardname=cardname, id=eid)
        for cardname, ix in self._ixs.items():
            if cardname == "GRID":
                continue
            bbox = self.get_bbox(cardname=cardname, id=eid)
            if bbox is not None:
                return bbox

    def update_grid(self, gid, /, old_coords, coords, rtree=None):
        return self._update(
            "GRID", gid, old_coords=old_coords, coords=coords, rtree=rtree
        )

    def update_element(self, eid, /, old_coords, coords, rtree=None):
        # old_coords is taken from
        try:
            cardname = self.reg.mesh.eid2card(skipcache=True)[eid]
        except:
            # breakpoint()
            raise
        return self._update(
            cardname, eid, old_coords=old_coords, coords=coords, rtree=rtree
        )

    def build(self):
        # add all grids to objects
        grids = self.reg.container["bulk"]["GRID"].array[["ID", "X1", "X2", "X3"]]
        grids = np.lib.recfunctions.structured_to_unstructured(grids)
        for row in grids:
            self.add_grid(row[0], coords=row[1:])
        # ===============================================
        # add elements
        eid2bbox = self.reg.mesh.eid2bbox()
        cards = self.reg.container["summary"]["element"]
        for dim in self.index_elems:
            cards = self.reg.container["summary"][dim]
            for cardname in cards:
                card = self.reg.container["bulk"][cardname]
                self._ixs[cardname] = index.Index(properties=self._rtree_properties)
                for i, (eid, gids) in enumerate(card.eid2gids().items()):
                    bbox = eid2bbox.loc[eid]
                    self.add_elem(cardname, eid, coords=bbox)

    def nearest_grids(self, /, coords, num_results=1, astype=set, rtree=None):
        if not rtree:
            rtree = self._ixs["GRID"]
        return astype(rtree.nearest(coords, num_results))

    def nearest_elements(
        self, coords, num_results=1, astype=set, cardname=None, rtree=None
    ):
        if rtree:
            return astype(rtree.nearest(coords, num_results))
        if cardname:
            ix = self._ixs[cardname]
            return astype(ix.nearest(coords, num_results))
        # get all ids
        ids = {}
        for cardname, ix in self._ixs.items():
            if cardname == "GRID":
                continue
            ids[cardname] = astype(ix.nearest(coords, num_results))
        return ids



"""
NASTRAN Axis Cards Collection
"""
import logging
import re
from collections import defaultdict

from nastranio.cardslib import SimpleCard
from nastranio.decorators import axis


# ============================================================================
# axis
# ============================================================================
@axis
class CORD2R(SimpleCard):
    """
    Rectangular Coordinate System Definition, Form 2
    Defines a rectangular coordinate system using the coordinates of three points.


    ref: NX Nastran 12 Quick Reference Guide 12-24 (p.1508)
    """

    TABLE = """
| 1      | 2   | 3   | 4  | 5  | 6  | 7  | 8  | 9  | 10 |
|--------+-----+-----+----+----+----+----+----+----+----|
| CORD2R | CID | RID | A1 | A2 | A3 | B1 | B2 | B3 |    |
|        | C1  | C2  | C3 |    |    |    |    |    |    |
    """
    DEFAULT = {"RID": 0}


@axis
class CORD2S(SimpleCard):
    """
    Spherical Coordinate System Definition, Form 2
    Defines a spherical coordinate system using the coordinates of three points.


    ref: NX Nastran 12 Quick Reference Guide 12-26 (p.1510)
    """

    TABLE = """
| 1      | 2   | 3   | 4  | 5  | 6  | 7  | 8  | 9  | 10 |
|--------+-----+-----+----+----+----+----+----+----+----|
| CORD2S | CID | RID | A1 | A2 | A3 | B1 | B2 | B3 |    |
|        | C1  | C2  | C3 |    |    |    |    |    |    |
    """
    DEFAULT = {"RID": 0}


@axis
class CORD2C(SimpleCard):
    """
    Cylindrical Coordinate System Definition, Form 2
    Defines a cylindrical coordinate system using the coordinates of three points.


    ref: NX Nastran 12 Quick Reference Guide 12-21 (p.1505)
    """

    TABLE = """
| 1      | 2   | 3   | 4  | 5  | 6  | 7  | 8  | 9  | 10 |
|--------+-----+-----+----+----+----+----+----+----+----|
| CORD2C | CID | RID | A1 | A2 | A3 | B1 | B2 | B3 |    |
|        | C1  | C2  | C3 |    |    |    |    |    |    |
    """
    DEFAULT = {"RID": 0}


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



"""
NASTRAN property Cards collection
"""
import logging
import re
from collections import defaultdict
from pprint import pprint

import numpy as np

from nastranio.cardslib import (
    ComplexCard,
    RepeatedRowsCard,
    SimpleCard,
    SimpleCyclingCard,
)
from nastranio.decorators import cached_property, fem_property


@fem_property
class PCOMP(SimpleCyclingCard):
    """
    Layered Composite Element Property.
    Defines the properties of an n-ply composite material laminate.

    ref: NX Nastran 12 Quick Reference Guide 16-97 (p.2327)


    >>> pcomp = PCOMP()
    >>> pcomp.append_fields_list(
    ...              [      1001, None, 0.00361, None,  None,  None, None, None,  '+',
    ...               '+',  1002, 0.018,    0.0, 'YES', 1003, 0.339,  0.0, 'YES', '+',
    ...               '+',  1009, 0.018,    5.0, 'YES'])
    >>> pcomp.append_fields_list(
    ...              [      1002, None, 0.005, None,  None,  None, None, None,  '+',
    ...               '+',  1002, 0.018,    0.0, 'YES', 1003, 0.339,  None, 'YES', '+',
    ...               '+',  1009, 0.018,    5.0, 'YES'])
    >>> # check main data:
    >>> pprint(pcomp.carddata['main'])  # doctest: +NORMALIZE_WHITESPACE
    defaultdict(<class 'list'>,
                {'FT': [None, None],
                 'GE': [None, None],
                 'LAM': [None, None],
                 'NSM': [0.00361, 0.005],
                 'PID': [1001, 1002],
                 'SB': [None, None],
                 'TREF': [None, None],
                 'Z0': [None, None],
                 'pcomp_layupID': [0, 0]})
    >>> # check LAYUPs creation. Both cards come with the same layout values. Therefore
    >>> # only one LAYUP (made of three plies) is created:
    >>> pprint(pcomp.carddata['pcomp_layup'])  # doctest: +NORMALIZE_WHITESPACE
    [[{'MID': 1002, 'SOUT': 'YES', 'T': 0.018, 'THETA': 0.0},
      {'MID': 1003, 'SOUT': 'YES', 'T': 0.339, 'THETA': 0.0},
      {'MID': 1009, 'SOUT': 'YES', 'T': 0.018, 'THETA': 5.0}]]
    """

    DIM = "2d"
    REPEATED_DATA_NAME = "layup"
    TABLE = """
| 1     | 2    | 3  | 4      | 5     | 6      | 7    | 8      | 9     | 10 |
|-------+------+----+--------+-------+--------+------+--------+-------+----|
| PCOMP | PID  | Z0 | NSM    | SB    | FT     | TREF | GE     | LAM   |    |
| ""    | MID1 | T1 | THETA1 | SOUT1 | MID2   | T2   | THETA2 | SOUT2 |    |
| ""    | MID3 | T3 | THETA3 | SOUT3 | -etc.- |      |        |       |    |
"""
    DEFAULTS = {
        "NSM": 0.0,
        "Z0": None,
        "SB": None,
        "FT": None,
        "TREF": None,
        "GE": None,
        "LAM": None,
    }
    REPEATED_DEFAULTS = {"THETA": 0.0, "SOUT": "NO"}
    COMMENTS_KEY = "Property"

    @cached_property
    def thk(self):
        """return a dict of thicknesses"""
        # calculate thicknesses for all layups
        layup_thks = {}
        for layID, data in enumerate(self.carddata["pcomp_layup"]):
            ts = [ply["T"] for ply in data]
            layup_thks[layID] = sum(ts)
        # map to PIDS
        thks = {
            pid: layup_thks[lay]
            for pid, lay in zip(
                self.carddata["main"]["PID"], self.carddata["main"]["pcomp_layupID"]
            )
        }
        return {
            "data": np.array(list(thks.values())),
            "index": np.array(list(thks.keys())),
            "name": "pid2thk",
        }

    @cached_property
    def pid2mids(self):
        """This ovverides the default "pid2mids" defined in `cardslib` module"""
        ret = {}

        for ix, pid in enumerate(self.carddata["main"]["PID"]):
            ret[pid] = set()
            layup_id = self.carddata["main"]["pcomp_layupID"][ix]
            layups = self.carddata["pcomp_layup"][layup_id]
            for layup in layups:
                ret[pid].add(layup["MID"])
            ret[pid] = frozenset(ret[pid])
        return ret


@fem_property
class PBAR(SimpleCard):
    """
    Simple Beam Property.
    Defines the properties of a simple beam element (CBAR entry).

    ref: NX Nastran 12 Quick Reference Guide 16-32 (p.2262)
    """

    DIM = "1d"
    COMMENTS_KEY = "Property"
    TABLE = """
| 1    | 2   | 3   | 4   | 5  | 6  | 7  | 8   | 9  | 10 |
|------+-----+-----+-----+----+----+----+-----+----+----|
| PBAR | PID | MID | A   | I1 | I2 | J  | NSM |    |
| ""   | C1  | C2  | D1  | D2 | E1 | E2 | F1  | F2 |
| ""   | K1  | K2  | I12 |    |    |    |     |    |
"""

    DEFAULTS = {
        "K1": 0.0,
        "K2": 0.0,
        "A": 0.0,
        "I1": 0.0,
        "I2": 0.0,
        "I12": 0.0,
        "J": 0.0,
        "NSM": 0.0,
        "C1": 0.0,
        "D1": 0.0,
        "E1": 0.0,
        "F1": 0.0,
        "C2": 0.0,
        "D2": 0.0,
        "E2": 0.0,
        "F2": 0.0,
    }


@fem_property
class PSHEAR(SimpleCard):
    """
    SHEAR Panel Property
    Defines the propertires of a shear panel (CSHEAR entry).

    ref: NX Nastran 12 Quick Reference Guide 16-208 (p.2438)


    """

    DIM = "2d"
    COMMENTS_KEY = "Property"
    TABLE = """
| 1      | 2   | 3    | 4    | 5    | 6  | 7   | 8  | 9  | 10 |
|--------+-----+------+------+------+----+-----+----+----+----|
| PSHEAR | PID | MID  | T    | NSM  | F1 | F2  |    |    |    |
"""
    DEFAULTS = {
        "NSM": 0.0,
        "F1": 0.0,
        "F2": 0.0,
    }

    @cached_property
    def thk(self):
        """return a dict of thicknesses"""
        thks = {
            pid: thk
            for pid, thk in zip(
                self.carddata["main"]["PID"], self.carddata["main"]["T"]
            )
        }
        # thks -> {1: 0.09, 203: 1.9}
        return {
            "data": np.array(list(thks.values())),
            "index": np.array(list(thks.keys())),
            "name": "pid2thk",
        }


@fem_property
class PSHELL(SimpleCard):
    """
    Shell Element Property
    Defines the membrane, bending, transverse shear, and coupling properties of thin
    shell elements.

    ref: NX Nastran 12 Quick Reference Guide 16-210 (p.2440)
    """

    DIM = "2d"
    COMMENTS_KEY = "Property"
    TABLE = """
| 1      | 2   | 3    | 4    | 5    | 6        | 7    | 8    | 9   | 10 |
|--------+-----+------+------+------+----------+------+------+-----+----|
| PSHELL | PID | MID1 | T    | MID2 | 12I/T**3 | MID3 | TS/T | NSM |    |
|        | Z1  | Z2   | MID4 |      |          |      |      |     |    |
"""
    DEFAULTS = {
        "12I/T**3": None,
        "TS/T": None,
        "NSM": 0.0,
        "Z1": None,
        "Z2": None,
        "MID4": None,
    }

    @cached_property
    def thk(self):
        """return a dict of thicknesses"""
        thks = {
            pid: thk
            for pid, thk in zip(
                self.carddata["main"]["PID"], self.carddata["main"]["T"]
            )
        }
        # thks -> {1: 0.09, 203: 1.9}
        return {
            "data": np.array(list(thks.values())),
            "index": np.array(list(thks.keys())),
            "name": "pid2thk",
        }


@fem_property
class PROD(SimpleCard):
    """
    Rod Property.

    Defines the properties of a rod element (CROD entry).

    ref: NX Nastran 12 Quick Reference Guide 16-206 (p.2436)
    """

    DIM = "1d"
    COMMENTS_KEY = "Property"
    TABLE = """
    | 1    | 2   | 3   | 4 | 5 | 6 | 7   | 8 | 9 | 10 |
    |------+-----+-----+---+---+---+-----+---+---+----|
    | PROD | PID | MID | A | J | C | NSM |   |   |    |
    """

    DEFAULTS = {"C": 0, "NSM": 0}


@fem_property
class PBUSH(ComplexCard):
    """
    Generalized Spring-and-Damper Property
    Defines the nominal property values for a generalized spring-and-damper structural
    element.

    ref: NX Nastran 12 Quick Reference Guide 16-83 (p.2800)
    """

    DIM = "1d"
    COMMENTS_KEY = "Property"
    TABLE = """
    | 1     | 2   | 3     | 4   | 5   | 6   | 7   | 8   | 9   | 10 |
    |-------+-----+-------+-----+-----+-----+-----+-----+-----+----|
    | PBUSH | PID | "K"   | K1  | K2  | K3  | K4  | K5  | K6  |    |
    |       |     | "B"   | B1  | B2  | B3  | B4  | B5  | B6  |    |
    |       |     | "GE"  | GE1 | GE2 | GE3 | GE4 | GE5 | GE6 |    |
    |       |     | "RCV" | SA  | ST  | EA  | ET  |     |     |    |
    """

    DEFAULTS = {
        "C": 0,
        "NSM": 0,
        '"K"': "K",
        "K1": None,
        "K2": None,
        "K3": None,
        "K4": None,
        "K5": None,
        "K6": None,
        '"B"': "B",
        "B1": None,
        "B2": None,
        "B3": None,
        "B4": None,
        "B5": None,
        "B6": None,
        '"GE"': "GE",
        "GE1": None,
        "GE2": None,
        "GE3": None,
        "GE4": None,
        "GE5": None,
        "GE6": None,
        '"RCV"': "RCV",
        "SA": None,
        "ST": None,
        "EA": None,
        "ET": None,
    }

    def clean_sparams(self, sparams):
        """remove"""
        Kfields = set((f"K{i}" for i in range(1, 6)))
        Ks = set((sparams[Ki] for Ki in Kfields))
        if Ks == {None}:
            Kfields.add('"K"')
            for field in Kfields:
                sparams.pop(field)
        Bfields = set((f"B{i}" for i in range(1, 6)))
        Bs = set((sparams[Bi] for Bi in Bfields))
        if Bs == {None}:
            Bfields.add('"B"')
            for field in Bfields:
                sparams.pop(field)
        GEfields = set((f"GE{i}" for i in range(1, 6)))
        GEs = set((sparams[GEi] for GEi in GEfields))
        if GEs == {None}:
            GEfields.add('"GE"')
            for field in GEfields:
                sparams.pop(field)


@fem_property
class PBEAM(RepeatedRowsCard):
    """
    Beam Property
    Defines the properties of a beam element (CBEAM entry). This element may be
    used to model tapered beams.

    ref: NX Nastran 12 Quick Reference Guide 16-48 (p.2278)

    > If SO is “YESA” or “NO”, the third continuation entry, which contains the fields
    > C1 through F2, must be omitted. If SO is “YES”, the continuation for Ci, Di, Ei,
    > and Fi must be the next entry.

    > ...

    > The fourth and fifth continuation entries, which contain fields K1 through
    > N2(B), are optional and may be omitted if the default values are appropriate.
    """

    DIM = "1d"
    COMMENTS_KEY = "Property"
    TABLE = """
    | 1     | 2     | 3      | 4     | 5     | 6     | 7      | 8     | 9      | 10 |
    |-------+-------+--------+-------+-------+-------+--------+-------+--------+----|
    | PBEAM | PID   | MID    | A(A)  | I1(A) | I2(A) | I12(A) | J(A)  | NSM(A) |    |
    |       | C1(A) | C2(A)  | D1(A) | D2(A) | E1(A) | E2(A)  | F1(A) | F2(A)  |    |
    """

    REPEATED_ROWS_TABLE = """
    | 1 | 2  | 3    | 4  | 5  | 6  | 7   | 8  | 9   | 10 |
    |---+----+------+----+----+----+-----+----+-----+----|
    |   | SO | X/XB | A  | I1 | I2 | I12 | J  | NSM |    |
    |   | C1 | C2   | D1 | D2 | E1 | E2  | F1 | F2  |    |
    """

    TRAILING_ROWS_TABLE = """
    | 1 | 2     | 3     | 4     | 5     | 6      | 7      | 8     | 9     | 10 |
    |---+-------+-------+-------+-------+--------+--------+-------+-------+----|
    |   | K1    | K2    | S1    | S2    | NSI(A) | NSI(B) | CW(A) | CW(B) |    |
    |   | M1(A) | M2(A) | M1(B) | M2(B) | N1(A)  | N2(A)  | N1(B) | N2(B) |
    """

    REPEATED_ROWS_NAME = "stations"
    TRIGGER_REPEATED_ON = str
    SKIP_NEXT_ROW_ON = ("SO", ("YESA", "NO"))
    DEFAULTS = {"I12(A)": 0.0, "J(A)": 0.0, "NSM(A)": 0.0}


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from pathlib import Path

import click

from nastranio.readers.gmsh.study import Study


@click.command()
@click.argument("meshfile", type=click.Path(exists=True))
@click.argument("params", type=click.Path(exists=True))
@click.option("--overwrite/--no-overwrite", default=False)
def study(meshfile, params, overwrite):
    study = Study(meshfile, autobuild=False)
    study.load_user_params(params)
    study.build()
    study.run(exist_ok=overwrite)


@click.group()
@click.pass_context
def main(ctx):
    ctx.ensure_object(dict)


@main.command()
@click.argument("meshfile", type=click.Path(exists=True))
@click.option("-p", "params", type=click.Path(exists=True), default=None)
@click.option("-o", "output", type=click.Path(exists=False), default=None)
def msh2nas(meshfile, params, output):
    meshfile = Path(meshfile)
    study = Study(meshfile, autobuild=False)
    if params is not None:
        params = Path(params)
    study.load_user_params(params)
    study.build()
    if output is not None:
        output = Path(output)
        if output.suffix in (".nas", ".dat", ".bulk"):
            study.to_nastran(target=output)
        elif output.suffix in (".msh", ".mesh"):
            study.reg.mesh.to_gmsh(filename=output)
    study.run()



"""
NASTRAN Elements Cards collection
"""
import logging
import re
from collections import defaultdict
from itertools import chain
from pprint import pprint

import numpy as np
from numtools.vgextended import loc_array

from nastranio.cardslib import ComplexCard, SimpleCard, SimpleCyclingCard
from nastranio.constants import VTKShapes, shapes
from nastranio.decorators import element


@element("0d", shapes.VERTICE)
class CONM2(SimpleCard):
    """
    Concentrated Mass Element Connection, Rigid Body Form
    Defines a concentrated mass at a grid point.


    ref: NX Nastran 12 Quick Reference Guide 12-4 (p.1488)
    """

    TABLE = """
    | 1     | 2   | 3   | 4   | 5   | 6   | 7   | 8  | 9 | 10 |
    |-------+-----+-----+-----+-----+-----+-----+----+---+----|
    | CONM2 | EID | G   | CID | M   | X1  | X2  | X3 |   |    |
    |       | I11 | I21 | I22 | I31 | I32 | I33 |    |   |    |
    """
    DEFAULTS = {"CID": 0}

    GIDS_PATTERN = re.compile("^G$")


@element("1d", shapes.LINE)
class CROD(SimpleCard):
    """
    Rod Element Connection.
    Defines a tension-compression-torsion element.


    ref: NX Nastran 12 Quick Reference Guide 12-120 (p.1604)
    """

    TABLE = """
    | 1    | 2   | 3   | 4  | 5  | 6 | 7  | 8 | 9 | 10 |
    |------+-----+-----+----+----+---+----+---+---+----|
    | CROD | EID | PID | G1 | G2 |   | SM |   |   |    |
    """

    DEFAULTS = {"C": 0, "NSM": 0}


# ============================================================================
# RBE3/RBE2
# ============================================================================


def _cells(cls, nasgids=None):
    """
    return RBE2 and RBE3 element's cell definition.
    """
    gidnames = ["nbpoints", cls.gids_header[0], "G2"]  # ['GN', 'G2']
    cells = [[], [], []]
    eids = cls.carddata["main"][cls.EID_FIELDNAME]
    eid2gids = cls._eid2gids
    # ========================================================================
    # default nasgids and naseids
    # ========================================================================
    if nasgids is None:
        logging.warning("no gids nor eids passed as reference")
        nasgids = list(cls._eid2gids.values())
        nasgids = np.array(sorted(list(chain(*nasgids))))
        naseids = np.array(sorted(list(cls.eid2gids().keys())))
    # -------------------------------------------------------------------------
    # for each element in cards,
    # create a bunch of dummy 2D element per leg, iterating over RBE nodes
    # calculate offset such as we ensure we do not point on existing EID
    faked_eids = []
    # print(cls.eid2gids())
    for eid in eids:
        master_gid = cls.array[np.where(cls.array["EID"] == eid)][cls.gids_header[0]][0]
        slave_gids = eid2gids[eid].copy()
        slave_gids.remove(master_gid)
        nb_fake_elements = len(slave_gids) - 1
        # append number of nodes
        cells[0] += (1 + nb_fake_elements) * [2]
        # ------------------------------------------------------------------------
        # map VTK nodes to NASTRAN nodes
        for slave_gid in slave_gids:
            faked_eids.append(eid)
            gids = [master_gid, slave_gid]
            gids = np.array(gids)
            gix = loc_array(nasgids, gids)
            cells[1].append(gix[0])
            cells[2].append(gix[1])

    payload = {
        "data": np.array(cells).T,
        "index": np.array(faked_eids),
        "columns": gidnames,
    }
    return payload


@element("1d", shapes.MPC)
class RBE3(ComplexCard):
    """
    Interpolation Constraint Element.

    Defines the motion at a reference grid point as the weighted average
    of the motions at a set of other grid points.

    ref: NX Nastran 12 Quick Reference Guide 17-48 (p.2522)
    """

    TABLE = """
    | 1    | 2       | 3     | 4       | 5    | 6    | 7      | 8    | 9      | 10 |
    |------+---------+-------+---------+------+------+--------+------+--------+----|
    | RBE3 | EID     |       | REFGRID | REFC | WT1  | C1     | G1,1 | G1,2   |    |
    |      | -etc.-  | WT2   | C2      | G2,1 | G2,2 | -etc.- | WT3  | C3     |    |
    |      | G3,1    | G3,2  | -etc.-  | WT4  | C4   | G4,1   | G4,2 | -etc.- |    |
    |      | "UM"    | GM1   | CM1     | GM2  | CM2  | GM3    | CM3  |        |    |
    |      |         | GM4   | CM4     | GM5  | CM5  | -etc.- |      |        |    |
    |      | "ALPHA" | ALPHA |         |      |      |        |      |        |    |
    """

    GIDS_PATTERN = re.compile("^REFGRID$")

    def __init__(self, name=None, data=None):
        super().__init__(name, data=data)
        # change self.fields and self.repeated
        self.fields = {fid: fname for fid, fname in self.fields.items() if fid <= 5}
        self.repeated = None
        if "rbe3_wcg" not in self.carddata:
            self.carddata["rbe3_wcg"] = []
        if "rbe3_um" not in self.carddata:
            self.carddata["rbe3_um"] = []

    def eid2gids_complement(self, eid=None, ix=None):
        """retrieve nodes stored in rbe3_wcg data"""
        gids = set()
        for data in self.carddata["rbe3_wcg"][ix]:
            gids |= data["Gi"]
        return gids

    def append_checkin(self, fields):
        """hook triggered right before `append`"""
        # fixed fields go to (including) field#5 "REFC"
        fixed_fields = fields[:4]  # to be parsed as usual
        self._remaining_fields = fields[4:]
        return fixed_fields

    def append_checkout(self, fields):
        """hook triggered right after `append`"""
        WCG_sequences = []
        UM_sequences = []
        buffer = None
        # we begin with WCG sequence
        FLAG = "WCG"
        _sequence = WCG_sequences
        fields = ["_", "_", "_", "_", "_", "_"] + self._remaining_fields

        for field_nb, field in enumerate(fields):
            try:
                next_field = fields[field_nb + 1]
            except:
                next_field = "EoF"
            if field == "_":
                continue
            # ----------------------------------------------------------------
            # skip continuation fields (fields #X0 or fields#X1
            if field_nb % 10 == 0 or (field_nb - 1) % 10 == 0:
                continue
            # ================================================================
            # WCG sequence
            # ================================================================
            if FLAG == "WCG":
                # ------------------------------------------------------------
                # opening a new sub-sequence, and closing the previous one
                # when we meet a float
                if isinstance(field, float):
                    if buffer:
                        # eventually close previous sub-sequence
                        _sequence.append(buffer)
                    buffer = []
                # ------------------------------------------------------------
                # closing the WCG sequence:
                #   * first None field parsed  (end of WCG, but not end of fields)
                #   * meeting "UM" or "ALPHA" flag
                #   * when reaching end of fields (next_field=='EoF')
                if field in ("UM", "ALPHA", None) or next_field == "EoF":
                    if next_field == "EoF":
                        # current field is still good
                        buffer.append(field)
                    # end of WCG current sequence AND also of WCG sequences
                    if buffer:
                        _sequence.append(buffer)
                        buffer = []
                    if field == "UM":
                        FLAG = "UM"
                        _sequence = UM_sequences
                    if field == "ALPHA":
                        FLAG = "ALPHA"
                    continue
                buffer.append(field)
                continue
            # ================================================================
            # UM sequence
            # ================================================================
            elif FLAG == "UM":
                # ------------------------------------------------------------
                # opening a new UM subsequence when buffer's length is two
                if len(buffer) == 2:
                    _sequence.append(buffer)
                    buffer = []
                # field #x9, #x2 are empty, but still in UM sequence
                if field is None:
                    # we theoritically fushed the buffer above, so...
                    assert len(buffer) == 0
                    continue
                # ------------------------------------------------------------
                # closing a UM sequence:
                #   * meeting "ALPHA" flag
                #   * when reaching end of fields (next_field=='EoF')
                if next_field == "EoF":
                    if buffer:
                        if len(buffer) not in (0, 2):
                            assert len(buffer) == 1
                            buffer.append(field)
                        _sequence.append(buffer)
                        buffer = []
                        IS_ALPHA_DEFAULT = True
                        alpha = 0.0
                        break
                if field == "ALPHA":
                    # end of WCG current sequence AND also of WCG sequences
                    if buffer:
                        if len(buffer) != 2:
                            __import__("pdb").set_trace()
                        _sequence.append(buffer)
                        buffer = []
                    if field == "ALPHA":
                        FLAG = "ALPHA"
                    continue
                buffer.append(field)
                continue
            elif FLAG == "ALPHA":
                IS_ALPHA_DEFAULT = False
                alpha = field
                break
        else:
            if buffer:
                _sequence.append(buffer)
            # default value
            IS_ALPHA_DEFAULT = True
            alpha = 0.0
        self.carddata["main"]["ALPHA"].append(alpha)
        # --------------------------------------------------------------------
        # process WCG sequence
        _all_wcg = []
        for weight, dof, *grids in WCG_sequences:
            _all_wcg.append({"W": weight, "C": dof, "Gi": set(grids)})
        if _all_wcg in self.carddata["rbe3_wcg"]:
            ix = self.carddata["rbe3_wcg"].index(_all_wcg)
        else:
            ix = len(self.carddata["rbe3_wcg"])
            self.carddata["rbe3_wcg"].append(_all_wcg)
        self.carddata["main"]["rbe3_wcgID"].append(ix)
        # --------------------------------------------------------------------
        # process UM sequence
        # print(UM_sequences)
        if UM_sequences in self.carddata["rbe3_um"]:
            ix = self.carddata["rbe3_um"].index(UM_sequences)
        else:
            ix = len(self.carddata["rbe3_um"])
            self.carddata["rbe3_um"].append(UM_sequences)
        self.carddata["main"]["rbe3_umID"].append(ix)
        return fields

    def cells(self, nasgids=None):
        """
        return element's cell definition.
        """
        return _cells(self, nasgids)


@element("1d", shapes.MPC)
class RBE2(SimpleCyclingCard):
    """
    Rigid Body Element, Form 2
    Defines a rigid body with independent degrees-of-freedom that are specified at a
    single grid point and with dependent degrees-of-freedom that are specified at an
    arbitrary number of grid points.

    ref: NX Nastran 12 Quick Reference Guide 17-45 (p.2519)

    >>> pb = RBE2()
    >>> pb.append_fields_list([      9, 8, 12, 10, 12, 14, 15, 16, '+',
    ...             '+', 20])
    >>> pb.append_fields_list([     10, 9, 12345, 10, 12, 14, 15, 16, '+',
    ...             '+', 100, 101, 5.])
    >>> pb.append_fields_list([     11, 109, 1245, 10, 12, 14, 15, 16, '+',
    ...             '+', 100, 101])  # no alpha specified. Expect default 0.0
    >>> pprint(pb.carddata['main'])  # doctest: +NORMALIZE_WHITESPACE
    defaultdict(<class 'list'>,
                {'ALPHA': [0.0, 5.0, 0.0],
                 'CM': [12, 12345, 1245],
                 'EID': [9, 10, 11],
                 'GN': [8, 9, 109],
                 'rbe2_gidsetID': [0, 1, 1]})
    >>> pprint(pb.export_data()['rbe2_gidset'])  # doctest: +NORMALIZE_WHITESPACE
    [[{'GM': 10}, {'GM': 12}, {'GM': 14}, {'GM': 15}, {'GM': 16}, {'GM': 20}],
     [{'GM': 10},
      {'GM': 12},
      {'GM': 14},
      {'GM': 15},
      {'GM': 16},
      {'GM': 100},
      {'GM': 101}]]

    """

    REPEATED_DATA_NAME = "gidset"
    TABLE = """
| 1    | 2   | 3   | 4   | 5      | 6     | 7   | 8   | 9   | 10 |
|------+-----+-----+-----+--------+-------+-----+-----+-----+----|
| RBE2 | EID | GN  | CM  | GM1    | GM2   | GM3 | GM4 | GM5 |    |
|      | GM6 | GM7 | GM8 | -etc.- | ALPHA |     |     |     |    |
    """
    DEFAULTS = {"ALPHA": 0.0}
    GIDS_PATTERN = re.compile("^GN$")

    def eid2gids_complement(self, eid=None, ix=None):
        # also include linked nodes
        try:
            compdata = self.carddata[self.REPEATED_DATA_NAME][ix]
        except IndexError as exc:
            logging.error(f"{ix=} out of bounds for {self.REPEATED_DATA_NAME=}")
            raise
        compdata = set([gid for subd in compdata for _, gid in subd.items()])
        return compdata
        # for subset_id, nodes in enumerate(subset):
        #     gidset_nodes[gidset].add(nodes['GM'])
        # eid2gidset = dict(zip(self.carddata['main'][self.EID_FIELDNAME], self.carddata['main']['rbe2_gidsetID']))
        # return eid2gidset

    def nb_nodes(self):
        return 2

    def cells(self, nasgids=None):
        """
        return element's cell definition.
        """
        return _cells(self, nasgids)


@element("2d", shapes.TRIA)
class CTRIA3(SimpleCard):
    """
    Triangular Plate Element Connection
    Defines an isoparametric membrane-bending or plane strain triangular plate element.

    ref: NX Nastran 12 Quick Reference Guide 12-147 (p.1631)
    """

    TABLE = """
    | 1      | 2   | 3     | 4  | 5  | 6  | 7          | 8     | 9 | 10 |
    |--------+-----+-------+----+----+----+------------+-------+---+----|
    | CTRIA3 | EID | PID   | G1 | G2 | G3 | THETA/MCID | ZOFFS |   |    |
    |        |     | TFLAG | T1 | T2 | T3 |            |       |   |    |
    """
    DEFAULTS = {
        "THETA/MCID": 0.0,
        "ZOFFS": None,
        "TFLAG": None,
        "T1": None,
        "T2": None,
        "T3": None,
    }
    MULT_TYPES_FIELDS = {"THETA/MCID": {float: "THETA", int: "MCID"}}


@element("2d", shapes.QUAD)
class CQUAD4(SimpleCard):
    """
    Quadrilateral Plate Element Connection

    Defines an isoparametric membrane-bending or plane strain quadrilateral plate element.

    ref: NX Nastran 12 Quick Reference Guide 12-80 (p.1564)

    >>> cq4 = CQUAD4()
    >>> bulk = ["CQUAD4         1       1       5      12      15      13",
    ...         "CQUAD4         2       4    4524    4537    4569    4547       4",
    ...         "CQUAD4         3       3       1       2       3       4      4."]
    >>> for b in bulk: cq4.parse(b)
    >>> pprint(cq4.export_data()['main']) # doctest: +NORMALIZE_WHITESPACE
    {'EID': [1, 2, 3],
     'G1': [5, 4524, 1],
     'G2': [12, 4537, 2],
     'G3': [15, 4569, 3],
     'G4': [13, 4547, 4],
     'MCID': [None, 4, None],
     'PID': [1, 4, 3],
     'T1': [None, None, None],
     'T2': [None, None, None],
     'T3': [None, None, None],
     'T4': [None, None, None],
     'TFLAG': [None, None, None],
     'THETA': [0.0, None, 4.0],
     'THETA/MCID': [0.0, 4, 4.0],
     'ZOFFS': [None, None, None]}
    """

    TABLE = """
| 1      | 2   | 3     | 4  | 5  | 6  | 7  | 8          | 9     | 10 |
|--------+-----+-------+----+----+----+----+------------+-------+----|
| CQUAD4 | EID | PID   | G1 | G2 | G3 | G4 | THETA/MCID | ZOFFS |    |
|        |     | TFLAG | T1 | T2 | T3 | T4 |            |       |    |
"""
    DEFAULTS = {
        "THETA/MCID": 0.0,
        "ZOFFS": None,
        "TFLAG": None,
        "T1": None,
        "T2": None,
        "T3": None,
        "T4": None,
    }
    MULT_TYPES_FIELDS = {"THETA/MCID": {float: "THETA", int: "MCID"}}


@element("2d", shapes.QUAD)
class CSHEAR(SimpleCard):
    """
    Shear Panel Element Connection

    Defines a Shear Panel element

    ref: NX Nastran 12 Quick Reference Guide 12-126 (p.1610)

    >>> cshear = CSHEAR()
    >>> bulk = ["CSHEAR         1       1       5      12      15      13",
    ...         "CSHEAR         2       4    4524    4537    4569    4547"]
    >>> for b in bulk: cshear.parse(b)
    >>> pprint(cshear.export_data()['main']) # doctest: +NORMALIZE_WHITESPACE
    {'EID': [1, 2],
     'G1': [5, 4524],
     'G2': [12, 4537],
     'G3': [15, 4569],
     'G4': [13, 4547],
     'PID': [1, 4]}
    """

    TABLE = """
| 1      | 2   | 3     | 4  | 5  | 6  | 7  | 8  | 9  | 10 |
|--------+-----+-------+----+----+----+----+----+----+----|
| CSHEAR | EID | PID   | G1 | G2 | G3 | G4 |    |    |    |
"""
    DEFAULTS = {}
    THK_PATTERN = None


@element("1d", shapes.LINE)
class CBUSH(SimpleCard):
    """
    Generalized Spring-and-Damper Connection
    Defines a generalized spring-and-damper structural element that may be nonlinear
    or frequency dependent.

    ref: NX Nastran 12 Quick Reference Guide 11-39 (p.1399)
    """

    TABLE = """
    | 1     | 2   | 3    | 4  | 5  | 6     | 7  | 8  | 9   | 10 |
    |-------+-----+------+----+----+-------+----+----+-----+----|
    | CBUSH | EID | PID  | GA | GB | GO/X1 | X2 | X3 | CID |    |
    |       | S   | OCID | S1 | S2 | S3    |    |    |     |    |
    """
    GIDS_PATTERN = re.compile("^G[A|B]$")
    DEFAULTS = {
        "S": None,
        "OCID": None,
        "S1": None,
        "S2": None,
        "S3": None,
        "CID": None,
    }
    MULT_TYPES_FIELDS = {"GO/X1": {float: "X1", int: "GO"}}


@element("1d", shapes.LINE)
class CBAR(SimpleCard):
    """
    Simple Beam Element Connection
    Defines a simple beam element.

    ref: NX Nastran 12 Quick Reference Guide 11-20 (p.1380)
    """

    TABLE = """
    | 1    | 2   | 3   | 4   | 5   | 6     | 7   | 8   | 9   | 10 |
    |------+-----+-----+-----+-----+-------+-----+-----+-----+----|
    | CBAR | EID | PID | GA  | GB  | GO/X1 | X2  | X3  |     |    |
    |      | PA  | PB  | W1A | W2A | W3A   | W1B | W2B | W3B |    |
    """
    MULT_TYPES_FIELDS = {"GO/X1": {float: "X1", int: "GO"}}
    GIDS_PATTERN = re.compile("^G[A|B]$")
    DEFAULTS = {
        "W1A": None,
        "W2A": None,
        "W3A": None,
        "W1B": None,
        "W2B": None,
        "W3B": None,
        "PA": None,
        "PB": None,
    }


@element("1d", shapes.LINE)
class CELAS2(SimpleCard):
    """
    Scalar Spring Property and Connection
    Defines a scalar spring element without reference to a property entry.

    ref: NX Nastran 12 Quick Reference Guide 11-77 (p.1437)
    """

    TABLE = """
    | 1      | 2   | 3 | 4  | 5  | 6  | 7  | 8  | 9 | 10 |
    |--------+-----+---+----+----+----+----+----+---+----|
    | CELAS2 | EID | K | G1 | C1 | G2 | C2 | GE | S |    |
    """


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



import logging
from functools import wraps

from numtools.intzip import hzip


class Modificator:
    """GmsHParser mixin class dedicated to modification"""

    def post_init(self):
        if not hasattr(self, "_locked_gids"):
            self._locked_gids = set()

    def lock_gids(self, gids):
        if isinstance(gids, (int, float)):
            gids = set((gids,))
        self._locked_gids |= set(gids)

    def renumber_nodes(self, /, mapping_in, lockin):
        """renumber nodes based on a mapping_in {current_node_id -> new_node_id}"""
        self.post_init()

        booked_gids = set(self.nodes)

        def _get_free_id():
            newgid = max(booked_gids) + 1
            booked_gids.add(newgid)
            return newgid

        # discard non-existing nodes from requirements
        mapping_in = {n1: n2 for n1, n2 in mapping_in.items() if n1 in self.nodes}
        logging.info(f"requested {mapping_in=}")
        # discard locked nodes
        forbidden = set(mapping_in) & self._locked_gids
        if forbidden:
            logging.warning(
                f"gids {hzip(forbidden)} are locked. They won't be renumbered"
            )
        mapping_in = {n1: n2 for n1, n2 in mapping_in.items() if n1 not in forbidden}
        # split dic_target into mapping_in , dic_back such
        buffer = {}
        temp_way_back = {}
        for n_source, n_target in mapping_in.items():
            intermediate_gid = _get_free_id()
            buffer[n_source] = intermediate_gid
            temp_way_back[intermediate_gid] = n_target
            if n_target in self.nodes:
                if n_target not in buffer:
                    intermediate_gid = _get_free_id()
                    buffer[n_target] = intermediate_gid
                    # temp_way_back[intermediate_gid] = intermediate_gid  # we may forget this one
                else:
                    pass
        way_back = {}
        for k, v in temp_way_back.items():
            if k not in buffer.values():
                continue
            way_back[k] = v
        assert sorted(list(set(way_back.values()))) == sorted(list(way_back.values()))
        assert sorted(list(set(buffer.values()))) == sorted(list(buffer.values()))
        mapping = {}
        for k, v in buffer.copy().items():
            mapping[k] = way_back.get(v, v)
            # if mapping[k] == k:
            #    buffer.pop(k, None)
            #    way_back.pop(k, None)
            #    mapping.pop(k, None)
        self._renumber_apply_mapping(buffer)
        self._renumber_apply_mapping(way_back)
        # provide appropriate executed mapping
        if lockin:
            to_lock = set(mapping_in.values())
            self._locked_gids |= to_lock
            logging.info(f"locked gids {hzip(to_lock)}")
        return mapping

    def _renumber_apply_mapping(self, dic):
        # reassign elements
        # print(f"apply {dic}")
        elts_to_modify = set()
        modified_elts = set()
        for n1 in dic:
            elts_to_modify = self.gid2eids[n1]
            # logging.info(f"node {n1} is used by {elts_to_modify}")
            for eid in elts_to_modify:
                elt = self.elements[eid]
                ix = elt.gids.index(n1)
                gids = list(elt.gids)
                gids[ix] = dic[n1]
                elt = elt._replace(gids=tuple(gids))
                self.elements[eid] = elt
                modified_elts.add(eid)
        # logging.info(f"modified elements {modified_elts}")
        # renumber nodes
        nodes_buffer = {}
        for gid, newgid in dic.items():
            node = self.nodes.pop(gid)
            node = node._replace(gid=newgid)
            nodes_buffer[newgid] = node
        self.nodes.update(nodes_buffer)
        self._make_gid_eid_type()
        self._make_eid_gid_physicalGroups()
        return modified_elts

    def renumber_nodes_by_grp(self, physical_name, startat=1, lockin=False):
        gids = self.physical_group2gids[physical_name]
        targets = list(range(startat, startat + len(gids) + 1))
        mapping = dict(zip(gids, targets))
        return self.renumber_nodes(mapping_in=mapping, lockin=lockin)

    def renumber_nodes_offset(self, offset, lockin=False):
        mapping = {k: k + offset for k in self.nodes}
        return self.renumber_nodes(mapping_in=mapping, lockin=lockin)



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

"""
Helpers to write NASTRAN cards
"""
from math import floor, log10


def nbrows_by_fields(fields):
    """return the number of rows that would be used to write provided
    fields dictionnary

    example: if two fields only are provided #17 and #23,
    we need to write row 11->20 and 21->30

    >>> nbrows_by_fields({17: 'robert', 23:'toto'})
    2
    """
    keys = fields.keys()
    kmax = max(keys)
    kmin = min(keys)
    # if kmin would be <=10, nb of rows would be
    rowid_min = (kmin - 1) // 10
    rowid_max = (kmax - 1) // 10
    return (rowid_max - rowid_min) + 1


def mantexp(f):
    """return the mantissa and exponent of a number as a tuple
    >>> mantexp(-2.3)
    (-2.3, 0)
    >>> mantexp(-1.123456789E-12)
    (-1.123456789, -12)
    """
    exponent = int(floor(log10(abs(f)))) if f != 0 else 0
    return f / 10**exponent, exponent


def _trans_float(value, field_length=8):
    """
    return float's string representation in NASTRAN way. Try to reduce as much
    as possible the string's length while keeping precision.


    >>> _trans_float(15)
    '15.'
    >>> _trans_float(-15.999)
    '-15.999'
    >>> _trans_float(15.9999999999999999999)
    '16.'
    >>> _trans_float(12345678)
    '1.2346+7'
    >>> _trans_float(123456789)
    '1.2346+8'
    >>> _trans_float(-0.123456789)
    '-.123457'
    >>> _trans_float(-0.999999999997388)
    '-1.'
    >>> _trans_float(0.999999999997388)
    '1.'
    >>> _trans_float(270000.0)
    '2.7000+5'
    """
    exponent = ""
    available_chars = field_length
    available_digits = field_length
    # ========================================================================
    # non-exponential numbers:
    #   * 0
    #   * range: ]-100000, -0.001]
    #   * range: [0.001, 100000[
    # ========================================================================
    # simple cases
    if value == 0:
        return "0."
    elif -1 < value < -0.001:
        # we loose two digits: "-."
        tpl = "{{:.{}f}}".format(field_length - 2)
        s = tpl.format(value)
        # return *minus* sign (s[0], skip leading "0" (s[1]) and trailing 0)
        if s[1] == "0":
            return s[0] + s[2:].rstrip("0")
        return s[0] + s[1:].rstrip("0")
    elif 0.001 < value < 1:
        # leading "0" can be omitted
        tpl = "{{:.{}f}}".format(field_length - 1)
        s = tpl.format(value)
        if s[0] == "0":
            return s[1:].rstrip("0")
        return s[0:].rstrip("0")

    # ------------------------------------------------------------------------
    # more complex stuff
    elif -100000 < value <= -1:
        # negative number. We loose one digit for the '-' sign and one for the '.'
        available_digits -= (
            len(str(int(value))) + 1
        )  # loose digits for integer part and dot
        tpl = "{{0:.{}f}}".format(available_digits)
        return tpl.format(value).rstrip("0")
    elif 1 <= value < 100000:
        available_digits -= (
            len(str(int(value))) + 1
        )  # loose digits for integer part and dot
        tpl = "{{0:.{}f}}".format(available_digits)
        return tpl.format(value).rstrip("0")
    # ========================================================================
    # exponential numbers:
    #   * range: -0.001, 0.001     # (excl. borns) small numbers
    #   * range: -inf,  -100000    # (incl. borns) big negative numbers
    #   * range: 100000, +inf      # (incl. borns) big negative numbers
    # ========================================================================
    else:
        # use exponent notation
        mantissa, exponent = mantexp(value)
        available_chars = field_length - len(str(exponent)) - 2
        if exponent > 0:
            available_chars -= 1
        if mantissa < 0:
            available_chars -= 1
        E_format = "{{mantissa:.{}f}}{{exponent:+d}}".format(available_chars)
        return E_format.format(mantissa=mantissa, exponent=exponent)


def _trans_int(val, field_length):
    """
    simply returns the integer as string (`str(val)`), except when resulting
    length is more than allowed length. In this latter case, an exception is raised.

    >>> _trans_int(1, 8)
    '1'
    """
    s = str(val)
    if len(s) <= field_length:
        return s
    # as per NASTRAN documentation, there shouldn't be any interger
    # whose length is longer than field's length
    raise ValueError(
        "encountered integer (%d) wider than %d chars" % (val, field_length)
    )


def trans(val, field_length=8):
    """translate field to NASTRAN compliant 8-characters fields

    >>> checks = ((6250, '6250'),
    ...           (0.0, '0.'),
    ...           (6250.0, '6250.'),
    ...           (-0.123456789, "-.123457"),
    ...           (0.123456789, ".1234568"),
    ...           (0.0023148, '.0023148'),
    ...           (-1.987654e-12, '-1.99-12'),
    ...           (None, ""))
    >>> err = []
    >>> for val, exp in checks:
    ...     if trans(val) != exp:
    ...         err.append((val, exp, trans(val)))
    >>> err
    []

    """
    if val is None:  # return blank field
        res = ""
    elif isinstance(val, float):
        res = _trans_float(val, field_length=field_length)
    elif isinstance(val, int):
        res = _trans_int(val, field_length=field_length)
    else:
        res = val
    # assert len(res) <= field_length
    return res


class DefaultDict(dict):
    def __missing__(self, key):
        return ""


def fields_to_card(fields, leading="", sep=""):
    """convert a single card dictionnary of fields to nastran card"""
    # the `fields` dict should look like:
    # {'fn1': str1, 'fn2': str2, ..., 'fn12': str3}
    if len(fields["fn1"]) + len(leading) + len(sep) > 8:
        raise ValueError('leading length is too long to fit with "%s"' % fields["fn1"])
    tpl = (
        "{leading}{{fn%d:{w1}}}{sep}{{fn%d:>{w}}}{sep}{{fn%d:>{w}}}{sep}{{fn%d:>{w}}}{sep}{{fn%d:>{w}}}{sep}"
        "{{fn%d:>{w}}}{sep}{{fn%d:>{w}}}{sep}{{fn%d:>{w}}}{sep}{{fn%d:>{w}}}{sep}{{fn%d:{w}}}{sep}"
    )
    _d = {
        "w1": 8 - len(leading) - len(sep),
        "w": 8 - len(sep),
        "leading": leading,
        "sep": sep,
    }
    tpl = tpl.format(**_d)
    # clean fields
    for fieldcode, fieldvalue in fields.copy().items():
        if not fieldvalue:
            fields.pop(fieldcode)
    # calculate the number of rows:
    fieldmax = max([int(k[2:]) for k in fields.keys()])
    nb_rows = fieldmax // 10 + 1
    # populate continuation fields
    for fnb in range(2, fieldmax):
        if fnb % 10 == 0 or (fnb - 1) % 10 == 0:
            fields["fn%d" % fnb] = "+"
    # prepare the multiline template
    tpls = []
    for i in range(0, nb_rows):
        ix = range(1 + i * 10, 11 + i * 10)
        tpls.append(tpl % tuple(ix))
    tpl = "\n".join(tpls)
    lines = tpl.format_map(fields).split("\n")
    lines = [l.strip() for l in lines]
    return lines


def get_field(field_nb):
    """simply return `field_nb` when `field_nb` is not a continuation field nb (10, 11,
    20, 21, etc.). Otherwise return next available field

    >>> get_field(1)
    1
    >>> get_field(11)
    12
    >>> get_field(14)
    14
    >>> get_field(10)
    12
    >>> get_field(11)
    12
    >>> get_field(21)
    22
    """
    if field_nb == 1:
        return 1
    if field_nb % 10 == 0:
        return field_nb + 2
    if (field_nb - 1) % 10 == 0:
        return field_nb + 1
    return field_nb


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)
